from collections import OrderedDict
from ..generic.general_methods import aedt_exception_handler, generate_unique_name
from ..application.DataHandlers import dict2arg, arg2dict
import copy

defaultparametricSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
            OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                                "StartingPoint": OrderedDict(), "Sim. Setups": [],
                                                "Sweeps": OrderedDict({"SweepDefinition":
                                                                           OrderedDict({"Variable": "",
                                                                                        "Data": "",
                                                                                        "OffsetF1": False,
                                                                                        "Synchronize": 0})}),
                                                "Sweep Operations": OrderedDict(),
                                                "Goals": OrderedDict()})


defaultdxSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
            OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                                "StartingPoint": OrderedDict(), "Sim. Setups": [],
                                                "Sweeps": OrderedDict({"SweepDefinition":
                                                                           OrderedDict({"Variable": "",
                                                                                        "Data": "",
                                                                                        "OffsetF1": False,
                                                                                        "Synchronize": 0})}),
                              "Sweep Operations": OrderedDict(), "CostFunctionName": "Cost", "CostFuncNormType": "L2",
                               "CostFunctionGoals": OrderedDict(),"EmbeddedParamSetup": -1, "Goals": OrderedDict()})

defaultoptiSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
    OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                "StartingPoint": OrderedDict(), "Optimizer": "Quasi Newton",
                                "AnalysisStopOptions": OrderedDict({"StopForNumIteration": True,
                                                                    "StopForElapsTime": False,
                                                                    "StopForSlowImprovement": False,
                                                                    "StopForGrdTolerance": False,
                                                                    "MaxNumIteration": 1000,
                                                                    "MaxSolTimeInSec": 3600,
                                                                    "RelGradientTolerance": 0,
                                                                    "MinNumIteration": 10
                                                                    }),
                                "CostFuncNormType": "L2", "PriorPSetup": "", "PreSolvePSetup": True,
                                "Variables": OrderedDict(), "LCS": OrderedDict(), "Goals": OrderedDict(),
                                "Acceptable_Cost": 0, "Noise": 0.0001, "UpdateDesign": False, "UpdateIteration": 5,
                                "KeepReportAxis": True, "UpdateDesignWhenDone": True})

defaultsensitivitySetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
    OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                       "StartingPoint": OrderedDict(), "MaxIterations": 10,
                                       "PriorPSetup": "", "PreSolvePSetup": True, "Variables": OrderedDict(),
                                       "LCS": OrderedDict(), "Goals": OrderedDict(), "Primary Goal": 0,
                                       "PrimaryError": 0.0001, "Perform Worst Case Analysis": False})

defaultstatisticalSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
    OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                       "StartingPoint": OrderedDict(), "MaxIterations": 50,
                                       "SeedValue": 0, "PriorPSetup": "", "Variables": OrderedDict(),
                                       "Goals": OrderedDict()})

defaultdoeSetup = OrderedDict({"IsEnabled": True, "ProdOptiSetupDataV2":
            OrderedDict({"SaveFields": False, "CopyMesh": False, "SolveWithCopiedMeshOnly": True}),
                                                "StartingPoint": OrderedDict(), "Sim. Setups": [],
                                                "CostFunctionName": "Cost",
                               "CostFuncNormType": "L2", "CostFunctionGoals": OrderedDict(), "Variables": OrderedDict(),
                               "Goals": OrderedDict(),
                               "DesignExprData": OrderedDict({"Type": "kOSF", "CCDDeignType": "kFaceCentered",
                                                              "CCDTemplateType": "kStandard",
                                                              "LHSSampleType": "kCCDSample",
                                                              "RamdomSeed": 0, "NumofSamples": 10,
                                                              "OSFDeignType": "kOSFD_MAXIMINDIST",
                                                              "MaxCydes": 10}),
                               "RespSurfaceSetupData": OrderedDict({"Type": "kGenAggr", "RefineType": "kManual"}),
                               "ResponsePoints": OrderedDict({"NumOfStrs": 0}),
                               "ManualRefinePoints": OrderedDict({"NumOfStrs": 0}),
                               "CustomVerifyPoints": OrderedDict({"NumOfStrs": 0}), "Tolerances": []})


class CommonOptimetrics(object):
    """Creates and sets up optimizations.
    
    Parameters
    ----------
    parent :
    
    name :
    
    dictinputs 
    
    optimtype : str
        Type of the optimization.
    
    """

    @property
    def omodule(self):
        """Optimetrics module.
        
        Returns
        -------
        :class:`Optimetrics`
            Optimetrics object.
        
        """
        return self._parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent, name, dictinputs, optimtype):
        self._parent = parent
        self.name = name
        self.soltype = optimtype

        inputd = copy.deepcopy(dictinputs)


        if optimtype == "OptiParametric":
            self.props = inputd or defaultparametricSetup
        if optimtype == "OptiDesignExplorer":
            self.props = inputd or defaultdxSetup
        if optimtype == "OptiOptimization":
            self.props = inputd or defaultoptiSetup
        if optimtype == "OptiSensitivity":
            self.props = inputd or defaultsensitivitySetup
        if optimtype == "OptiStatistical":
            self.props = inputd or defaultstatisticalSetup
        if optimtype == "OptiDXDOE":
            self.props = inputd or defaultdoeSetup

        if inputd:
            self.props.pop('ID', None)
            self.props.pop('NextUniqueID', None)
            self.props.pop('MoveBackwards', None)
            self.props.pop('GoalSetupVersion', None)
            self.props.pop('Version', None)
            self.props.pop('SetupType', None)
            if inputd.get("Sim. Setups"):
                setups = inputd["Sim. Setups"]
                for el in setups:
                    if type(self._parent.design_properties["SolutionManager"]['ID Map']['Setup']) is list:
                        for setup in self._parent.design_properties["SolutionManager"]['ID Map']['Setup']:
                            if setup['I'] == el:
                                setups[setups.index(el)]= setup['I']
                                break
                    else:
                        if self._parent.design_properties["SolutionManager"]['ID Map']['Setup']['I'] == el:
                            setups[setups.index(el)] = self._parent.design_properties["SolutionManager"]['ID Map']['Setup']['N']
                            break
            if inputd.get("Goals", None):
                oparams = self.omodule.GetChildObject(self.name).GetCalculationInfo()
                oparam = [i for i in oparams[0]]
                calculation = ["NAME:Goal"]
                calculation.extend(oparam)
                arg1 = OrderedDict()
                arg2dict(calculation, arg1)
                self.props['Goals'] = arg1

    @aedt_exception_handler
    def update(self, update_dictionary=None):
        """Update the setup based on stored properties. 

        Parameters
        ----------
        update_dictionary : dict, optional
            Dictionary to use. The  default is ``None``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        if update_dictionary:
            for el in update_dictionary:
                self.props[el]=update_dictionary[el]

        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)

        self.omodule.EditSetup(self.name, arg)
        return True

    @aedt_exception_handler
    def create(self):
        """Create a setup.
        
        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        
        """
        arg = ["NAME:" + self.name]
        dict2arg(self.props, arg)
        self.omodule.InsertSetup(self.soltype, arg)
        return True

    @aedt_exception_handler
    def _add_calculation(self, reporttype, solution=None, domain="Sweep", calculation="", calculation_type="d",
                         calculation_value="", calculation_name=None):
        """Add a calculation to the setup.

        Parameters
        ----------
        reporttype : str
            Type of report. 
        solution : str, optional
            Type of the solution. The default is ``None``.
        domain : str, optional
             Type of the domain. The default is ``"Sweep"``.
        calculation : str, optional
             The default is ``""``.
        calculation_type : str, optional
             Type of the calculaton. The default is ``"d"``.
        calculation_value : str, optional
             The default is ``""``.
        calculation_name : str, optional
             Name of the the calculation. The default is ``None``.

        Returns
        -------

        """
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep

        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        if calculation_name:
            sweepdefinition["Name"] = calculation_name
        else:
            sweepdefinition["Name"] = generate_unique_name(calculation)
        if domain == "Sweep":
            var = "Freq"
        else:
            var ="Time"
        sweepdefinition["Ranges"] = OrderedDict(
            {"Range": ["Var:=", var, "Type:=", calculation_type, "DiscreteValues:=", calculation_value]})
        if "Goal" in self.props["Goals"]:
            if type(self.props["Goals"]["Goal"]) is not list:
                self.props["Goals"]["Goal"] = [self.props["Goals"]["Goal"], sweepdefinition]
            else:
                self.props["Goals"]["Goal"].append(sweepdefinition)
        else:
            self.props["Goals"]["Goal"] = sweepdefinition

        return self.update()

    @aedt_exception_handler
    def _add_goal(self, optigoalname, reporttype, solution=None, domain="Sweep", calculation="", calculation_type="discrete",
                  calc_val1="", calc_val2="", condition="==", goal_value=1, goal_weight=1, goal_name=None):
        """Add an optimization goal to the setup.

        Parameters
        ----------
        optigoalname : str
            Name of the optimization goal.
        reporttype : str, optional
            Type of the report.
        solution : str, optional
            Type of the solution. The default is ``None``.
        domain : str, optional
            Type of the domain. The default is ``"Sweep"''.
        calculation : str, optional
            Name of the calculation. The default is ``""``.
        calculation_type : str, optional
            Type of the calculation. The default is ``"discrete"``.
        calc_val1 : str, optional
            First value for the calculation. The default is ``""``.
        calc_val2 : str, optional
            Second value for the calculation. The default is ``""``.
        condition : str, optional
            The condition for the calculation. The default is ``"=="``.
        goal_value : optional
            Value for the optimization goal. The default is ``1``.
        goal_weight : optional
            Weight for the optimzation goal. The default is ``1``.
        goal_name : str, optional
            Name of the goal. The default is ``None``.

        Returns
        -------

        """
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep
        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        if goal_name:
            sweepdefinition["Name"] = goal_name
        else:
            sweepdefinition["Name"] = generate_unique_name(calculation)
        if domain == "Sweep":
            var = "Freq"
        else:
            var ="Time"
        if calculation_type =="discrete":
            if type(calc_val1) is list:
                dr = ",".join(calc_val1)
            else:
                dr = calc_val1
            sweepdefinition["Ranges"] = OrderedDict(
                {"Range": ["Var:=", var, "Type:=", "d", "DiscreteValues:=", dr]})
        else:
            sweepdefinition["Ranges"] = OrderedDict({"Range": ["Var:=", var, "Type:=", calculation_type,
                                                               "Start:=", calc_val1, "Stop:=", calc_val2,
                                                               "DiscreteValues:=", ""]})
        sweepdefinition["Condition"] = condition
        sweepdefinition["GoalValue"] = OrderedDict(
            {"GoalValueType": "Independent", "Format": "Real/Imag", "bG": ["v:=", "[{};]".format(goal_value)]})
        sweepdefinition["Weight"] = "[{};]".format(goal_weight)
        if "Goal" in self.props[optigoalname]:
            if type(self.props[optigoalname]["Goal"]) is not list:
                self.props[optigoalname]["Goal"] = [self.props[optigoalname]["Goal"], sweepdefinition]
            else:
                self.props[optigoalname]["Goal"].append(sweepdefinition)
        else:
            self.props[optigoalname]["Goal"] = sweepdefinition
        return self.update()


class DXSetups(object):
    """Sets up DesignXplorer optimizations.
    
    Parameters
    ----------
    parent :
        
    """

    class Setup(CommonOptimetrics, object):
        """Sets up a DesignXplorer optimization in optiSLang.

        Parameters
        ----------
        parent :
        name :
        dictinputs :
            The default is ``None``.
        optimtype : str, optional
            Type of the optimization. The default is ``"OptiDesignExplorer"``.
      
        """
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiDesignExplorer")

        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None, domain="Sweep",  calculation_name=None):
            """Add a calculation to the setup.

            Parameters
            ----------
            calculation : str, optional
                Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
            calculation_value : str, optional
                Value for the calculation, such as ``"1GHz"`` if a sweep. The default is ``""``.
            reporttype : str, optional
                Name of the report to add the calculation to. The default 
                is ``"Modal Solution Data"``.
            solution : str, optional
                Type of the solution. The default is ``None``, in which case the default 
                solution is used. 
            domain : str, optional
                Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
            calculation_name : str, optional
                 Name of the calculation. The default is ``None``.

            Returns
            -------
            bool
                ``True`` when successful, ``False`` when failed.
            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)

        @aedt_exception_handler
        def add_goal(self, calculation="", calculation_value="", calculation_type="discrete", calculation_stop="",
                     reporttype="Modal Solution Data", solution=None, domain="Sweep", goal_name=None, goal_value=1, goal_weight=1, condition="=="):
            """Add a goal to the setup.

            Parameters
            ----------
            calculation : str, optional
                Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
            calculation_value : str, optional
                Value for the calculation, such as ``"1GHz"`` if a sweep. The default is ``""``.
                If ``calculation_type="discrete"``, the value is discrete or is a list. If the
                value is a range, it is the starting value. 
            calculation_type : str, optional
                Type of the calculation. Options are ``"discrete"`` or ``"range"``. 
                The default is ``"discrete"``.
            calculation_stop : str, optional
                Stopping value for the calculation if ``calculation_type="range"``.  
                The default is ``""``.
            reporttype : str, optional
                Name of the report to add the calculation to. The default 
                is ``"Modal Solution Data"``.
            solution : str, optional
                Type of the solution. The default is ``None``, in which case the default 
                solution is used. 
            domain : str, optional
                Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
            goal_name : str, optional
                 Name of the goal. The default is ``None``.
            goal_value : optional
                 Value for the goal. The default is ``1``.
            goal_weight : optional
                 Value for the goal weight. The default is ``1``.
            condition : string, optional
                 The default is ``"=="``.

            Returns
            -------
            bool
                ``True`` when successful, ``False`` when failed.
            
            """
            return self._add_goal(optigoalname="CostFunctionGoals", reporttype=reporttype, solution=solution,
                                  domain=domain, calculation_type=calculation_type, calculation=calculation,
                                  calc_val1=calculation_value, calc_val2=calculation_stop, goal_name=goal_name,
                                  goal_weight=goal_weight, goal_value=goal_value, condition=condition)
    @property
    def parent(self):
        """Parent object."""
        return self._parent
    
    @property
    def optimodule(self):
        """Optimetrics module.
        
        Returns
        :class:`Optimetrics`
            Optimetrics object.
        
        """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiDesignExplorer":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_dx_setup(self,variables_to_include, defaults_var_values=None, setupname=None, parametricname=None):
        """Add a basic parametric setup in DesignXplorer.
        
        You can customize all DesignXplorer options after the setup is added.

        Parameters
        ----------
        variables_to_include : list
            List of variables to include in DesignXplorer.
        defaults_var_values : list, optional
            List of default variable values.
        setupname : str, optional
            Name of the setup. The default is ``None``, in which case the default 
            analysis setup is used.
        parametricname : str, optional
            Name of the parametric setup. The default is ``None``.

        Returns
        -------
        :class:`Optimetrics`
            Optimetrics object.
        
        """
        if not setupname:
            setupname = [self._parent.analysis_setup]
        elif type(setupname) is not list:
            setupname = [setupname]
        if not parametricname:
            parametricname = generate_unique_name("DesignXplorer")
        setup = self.Setup(self._parent, parametricname)
        setup.props["Sim. Setups"] = setupname
        setup.props['Sweeps'] = []
        if not defaults_var_values:
            for v in variables_to_include:
                sweepdefinition = OrderedDict()
                sweepdefinition["Variable"] = v
                if "$" in v:
                    sweepdefinition["Data"] = self._parent.oproject.GetVariableValue(v)
                else:
                    sweepdefinition["Data"] = self._parent.odesign.GetVariableValue(v)
                sweepdefinition["OffsetF1"] = False
                sweepdefinition["Synchronize"] = 0
        else:
            for v,vv in zip(variables_to_include, defaults_var_values):
                sweepdefinition = OrderedDict()
                sweepdefinition["Variable"] = v
                sweepdefinition["Data"] = vv
                sweepdefinition["OffsetF1"] = False
                sweepdefinition["Synchronize"] = 0
                setup.props['Sweeps'].append(sweepdefinition)
                setup.props["StartingPoint"][v] = vv
        setup.create()
        self.setups.append(setup)
        return setup


class ParametericsSetups(object):
    """Sets up a parametric analysis.
    
    Parameters
    ----------
    parent :
    
    """
    class Setup(CommonOptimetrics, object):

        """Sets up a parametric analysis in optiSLang.
        
        Parameters
        ----------
        parent :
        
        name :
        
        dictinputs : optional
            The default is ``None``.
        otimtype : str, optional
            Type of the optimization. The default is ``"OptiParametric"``.
        
        """
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiParametric")
            pass

        @aedt_exception_handler
        def add_variation(self, sweep_var, datarange):
            """Add a variation to an existing parametric setup.

            Parameters
            ----------
            sweep_var : str
                Name of the variable.
            datarange :
                Range of the data. 

            Returns
            -------
            bool
                ``True`` when successful, ``False`` when failed.
            
            """
            if type(self.props["Sweeps"]["SweepDefinition"]) is not list:
                self.props["Sweeps"]["SweepDefinition"] = [self.props["Sweeps"]["SweepDefinition"]]
            sweepdefinition = OrderedDict()
            sweepdefinition["Variable"] = sweep_var
            sweepdefinition["Data"] = datarange
            sweepdefinition["OffsetF1"] = False
            sweepdefinition["Synchronize"] = 0
            self.props["Sweeps"]["SweepDefinition"].append(sweepdefinition)
            self.update()
            return True

        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None,
                            domain="Sweep", calculation_name=None):
            """Add a calculation to the parametric setup.
            
            Parameters
            ----------
            calculation : str, optional
                Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
            calculation_value : str, optional
                Value for the calculation, such as ``"1GHz"`` if a sweep. The default is ``""``.
            reporttype : str, optional
                Name of the report to add the calculation to. The default 
                is ``"Modal Solution Data"``.
            solution : str, optional
                Type of the solution. The default is ``None``, in which case the default 
                solution is used. 
            domain : str, optional
                Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
            calculation_name : str, optional
                Name of the calculation. The default is ``None``.

            Returns
            -------
            bool
                ``True`` when successful, ``False`` when failed.
            
            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)

    @property
    def parent(self):
        """Parent object."""
        return self._parent

    @property
    def optimodule(self):
        """Optimetrics module.
        
        Returns
        -------
        :class:`Optimetrics`
            Optimetrics object.
        """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]

                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiParametric":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_parametric_setup(self, sweep_var, datarange, setupname=None, parametricname=None):
        """Add a basic parametric setup. 
        
        You can customize all options after the parametric setup is added.

        Parameters
        ----------
        sweep_var : str
            Name of the sweep.
        datarange :
            Range of the data.
        setupname : str, optional
            Name of the setup. The default is ``None``, in which case  the default 
            parametric setup is used.
        parametricname : str, optional
            Name of the parametric setup. The default is ``None``.

        Returns
        -------
        :class:`Optimetrics`
            Optimetrics object.
        
        """
        if not setupname:
            setupname = [self._parent.analysis_setup]
        elif type(setupname) is not list:
            setupname = [setupname]
        if not parametricname:
            parametricname = generate_unique_name("Parametric")
        setup = self.Setup(self._parent, parametricname)
        setup.props["Sim. Setups"] = setupname
        sweepdefinition = OrderedDict()
        sweepdefinition["Variable"] = sweep_var
        sweepdefinition["Data"] = datarange
        sweepdefinition["OffsetF1"] = False
        sweepdefinition["Synchronize"] = 0
        setup.props["Sweeps"]["SweepDefinition"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup


class SensitivitySetups(object):
    """Sets up a sensitivity analysis.
    
    Parameters
    ----------
    parent :
    
    """
    class Setup(CommonOptimetrics, object):
        """Sets up a sensitivity analysis in optiSLang.
        
        Parameters
        ----------
        parent :
        
        name :
        
        dictinputs : optional
            The default is ``None``.
        otimtype : str, optional
            Type of the optimization. The default is ``"OptiSensitivity"``.
        
        """
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiSensitivity")

        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None,
                            domain="Sweep", calculation_name=None):
            """Add a calculation to the sensitivity analysis.

            Parameters
            ----------
            calculation : str, optional
                Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
            calculation_value : str, optional
                Value for the calculation, such as ``"1GHz"`` if a sweep. The default is ``""``.
            reporttype : str, optional
                Name of the report to add the calculation to. The default 
                is ``"Modal Solution Data"``.
            solution : str, optional
                Type of the solution. The default is ``None``, in which case the default 
                solution is used. 
            domain : str, optional
                Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
            calculation_name : str, optional
                Name of the calculation. The default is ``None``.

            Returns
            -------
            bool
                ``True`` when successful, ``False`` when failed.
            
            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)

    @property
    def parent(self):
        """Parent object."""
        return self._parent
   
    @property
    def optimodule(self):
        """Optimetrics module.
        
        Returns
        -------
        :class:`Optimetrics`
            Optimetrics object.
            
        """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiSensitivity":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_sensitivity(self, calculation, calculation_value, calculation_type="Freq",
                        reporttype="Modal Solution Data", domain="Sweep", solution=None, parametricname=None):
        """Add a basic sensitivity analysis. 
        
        You can customize all options after the analysis is added.

        Parameters
        ----------
        calculation : str, optional
            Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
        calculation_value : str, optional
            Value for the calculation, such as ``"1GHz"`` if a sweep. The default is ``""``.
        calculation_type : str, optional
            Type of variation. The default is ``"Freq"``.
        reporttype : str, optional
            Name of the report to add the calculation to. The default 
            is ``"Modal Solution Data"``.
        domain : str, optional
            Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
        solution : str, optional
            Type of the solution. The default is ``None``, in which case the default 
            solution is used. 
        parametricname : str, optional
            Name of the sensitivity analysis. The default is ``None``, in which case 
            a default name is assigned.

        Returns
        -------
        :class:`Sensitivity`
            Sensitivity object.

        """
        if not parametricname:
            parametricname = generate_unique_name("Sensitivity")
        setup = self.Setup(self._parent, parametricname)
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep
        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        sweepdefinition["Name"] = calculation
        sweepdefinition["Ranges"] = OrderedDict(
            {"Range": ["Var:=", calculation_type, "Type:=", "d", "DiscreteValues:=", calculation_value]})
        setup.props["Goals"]["Goal"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup


class StatisticalSetups(object):
    """Sets up a statistical analysis.
    
    Parameters
    ----------
    parent :
    
    """
    class Setup(CommonOptimetrics, object):
        """Sets up a statistical analysis in optiSLang.
        
        Parameters
        ----------
        parent :
        
        name :
        
        dictinputs : optional
            The default is ``None``.
        otimtype : str, optional
            Type of the optimization. The default is ``"OptiStatistical"``.
        
        """
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiStatistical")
            pass

        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None, domain="Sweep",  calculation_name=None):
            """Add a calculation to the statistical analysis.

            Parameters
            ----------
            calculation : str, optional
                Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
            calculation_value : str, optional
                Value for the calculation, such as ``"1GHz"`` if a sweep. The default is ``""``.
            reporttype : str, optional
                Name of the report to add the calculation to. The default 
                is ``"Modal Solution Data"``.
            domain : str, optional
                Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
            solution : str, optional
                Type of the solution. The default is ``None``, in which case the default 
                solution is used. 
            calculation_name : str, optional
                Name of the calculation. The default is ``None``.

            Returns
            -------
            bool
                ``True`` when successful, ``False`` when failed.
            
            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)

    @property
    def parent(self):
        """Parent object."""
        return self._parent
    
    @property
    def optimodule(self):
        """Optimetrics module.
        
        Returns
        -------
        :class:`Optimetrics`
            Optimetrics object.
            
        """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiStatistical":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_statistical(self, calculation_name, calc_variation_value, calculation_type="Freq",
                        reporttype="Modal Solution Data", domain="Sweep", solution=None, parametricname=None):
        """Add a basic statistical analysis. 
        
        You can customize all options after the analysis is added.

        Parameters
        ----------
        calculation_name : str, optional
            Name of the calculation.
        calc_variation_value : str, optional
            Variation value, such as ``"1GHz"``.
        calculation_type : str, optional
            Variation type. The default is ``"Freq"``.
        reporttype : str, optional
            Type of report to add the calculation to. The default is 
            ``"Modal Solution Data"``.
        domain : str, optional
            Type of the domain. The default is ``"Sweep"``.
        solution : str, optional
            Name of the solution. The default is ``None``, in which case the
            nominal sweep is used.
        parametricname : str, optional
            Name of the analysis. The default is ``None``, in which case the
            default name is assigned.

        Returns
        -------
        :class:`Statistical`
            Statistical object.
     
        """
        if not parametricname:
            parametricname = generate_unique_name("Statistical")
        setup = self.Setup(self._parent, parametricname)
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep

        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation_name
        sweepdefinition["Name"] = calculation_name
        sweepdefinition["Ranges"] = OrderedDict({"Range": ["Var:=", calculation_type,	"Type:=", "d", "DiscreteValues:=", calc_variation_value]})
        setup.props["Goals"]["Goal"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup


class DOESetups(object):
    """Sets up DOEs (Designs of Experiments).
    
    Parameters
    ----------
    parent :
   
    """
    class Setup(CommonOptimetrics, object):
        """Sets up a DOE (Design of Experiments) in optiSLang.
        
        Parameters
        ----------
        parent :
        
        name :
        
        dictinputs : optional
            The default is ``None``.
        otimtype : str, optional
            Type of the optimization. The default is ``"OptiDXDOE"``.
        
        """
        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiDXDOE")
            pass
        @aedt_exception_handler
        def add_calculation(self, calculation="", calculation_value="", reporttype="Modal Solution Data", solution=None,
                            domain="Sweep", calculation_name=None):
            """Add a calculation to the DOE (Design of Experiments).

            Parameters
            ----------
            calculation : str, optional
               Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
            calculation_value : str, optional
                Value for the calculation, such as ``"1GHz"`` if a sweep. The default is ``""``.
            reporttype : str, optional
                Name of the report to add the calculation to. The default 
                is ``"Modal Solution Data"``.
            solution : str, optional
                Type of the solution. The default is ``None``, in which case the default 
                solution is used. 
            domain : str, optional
                Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
            calculation_name : str, optional
                Name of the calculation. The default is ``None``.       

            Returns
            -------
            bool
                ``True`` when successful, ``False`` when failed.
            
            """
            return self._add_calculation(reporttype=reporttype, solution=solution, domain=domain, calculation_type="d",
                                         calculation=calculation, calculation_value=calculation_value, calculation_name=calculation_name)

        @aedt_exception_handler
        def add_goal(self, calculation="", calculation_value="", calculation_type="discrete",
                     calculation_stop="", reporttype="Modal Solution Data", solution=None, domain="Sweep", goal_name=None, goal_value=1, goal_weight=1, condition="=="):
            """Add a goal to the DOE (Design of Experiments).

            Parameters
            ----------
            calculation : str, optional
                Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
            calculation_value : str, optional
                Value for the calculation, such as ``"1GHz"`` if a sweep. The default is ``""``.
                If ``calculation_type="discrete"``, the value is discrete or is a list. If the
                value is a range, it is the starting value. 
            calculation_type : str, optional
                Type of the calculation. Options are ``"discrete"`` or ``"range"``. 
                The default is ``"discrete"``.
            calculation_stop : str, optional
                Stopping value if ``calculation_type="range"``.  The default is ``""``.
            reporttype : str, optional
                Name of the report to add the calculation to. The default 
                is ``"Modal Solution Data"``.
            solution : str, optional
                Type of the solution. The default is ``None``, in which case the default 
                solution is used. 
            domain : str, optional
                Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
            goal_name : str, optional
                 Name of the goal. The default is ``None``.
            goal_value : optional
                 Value for the goal. The default is ``1``.
            goal_weight : optional
                 Value for the goal weight. The default is ``1``.
            condition : string, optional
                 The default is ``"=="``.

            Returns
            -------
            bool
                ``True`` when successful, ``False`` when failed.
            
            """
            return self._add_goal(optigoalname="CostFunctionGoals", reporttype=reporttype, solution=solution,
                                  domain=domain, calculation_type=calculation_type, calculation=calculation,
                                  calc_val1=calculation_value, calc_val2=calculation_stop, goal_name=goal_name,
                                  goal_weight=goal_weight,
                                  goal_value=goal_value, condition=condition)

    @property
    def parent(self):
        """Parent object."""
        return self._parent

    @property
    def optimodule(self):
        """Optimetrics module.
        
        Returns
        -------
        :class:`Optimetrics`
            Optimetrics object.
        
        """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiDXDOE":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass

    @aedt_exception_handler
    def add_doe(self, calculation, calculation_value, calculation_type="Freq", reporttype="Modal Solution Data",
                domain="Sweep", condition="<=", goal_value=1, goal_weight=1, solution=None, parametricname=None):
        """Add a basic DesignXplorer DOE (Design of Experiments). 
        
        You can customize all options after the DOE is added.

        Parameters
        ----------
        calculation : str, optional
            Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
        calculation_value : str, optional
            Variation value, such as ``"1GHz"``.
        calculation_type : str, optional
            Type of the calculation. The default is ``"Freq"``.
        calculation_stop : str, optional
            Stopping value if ``calculation_type="range"``.  The default is ``""``.
        reporttype : str, optional
            Name of the report to add the calculation to. The default 
            is ``"Modal Solution Data"``.
        domain : str, optional
            Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
        condition : string, optional
            The default is ``"<="``.
        goal_value : optional
            Value for the goal. The default is ``1``.
        goal_weight : optional
            Value for the goal weight. The default is ``1``.
        solution : str, optional
            Type of the solution. The default is ``None``, in which case the default 
            solution is used.
        parametricname : str, optional
            Name of the analysis. The default is ``None``, in which case a
            default name is assigned.

        Returns
        -------
        type
            DOE object.

        """
        if not solution:
            solution = self._parent.nominal_sweep
        setupname = [solution.split(" ")[0]]
        if not parametricname:
            parametricname = generate_unique_name("DesignOfExp")
        setup = self.Setup(self._parent, parametricname)
        setup.props["Sim. Setups"] = setupname
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype

        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        sweepdefinition["Name"] = calculation
        sweepdefinition["Ranges"] = OrderedDict(
            {"Range": ["Var:=", calculation_type, "Type:=", "d", "DiscreteValues:=", calculation_value]})
        sweepdefinition["Condition"] = condition
        sweepdefinition["GoalValue"] = OrderedDict(
            {"GoalValueType": "Independent", "Format": "Real/Imag", "bG": ["v:=", "[{};]".format(goal_value)]})
        sweepdefinition["Weight"] = "[{};]".format(goal_weight)
        setup.props["CostFunctionGoals"]["Goal"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup


class OptimizationSetups(object):
    """Sets up optimizations.
    
    Parameters
    ----------
    parent :
        
    """

    class Setup(CommonOptimetrics, object):
        """Sets up an optimization in optiSLang.
        
        Parameters
        ----------
        parent :
        
        name :
        
        dictinputs : optional
            The default is ``None``.
        
        """

        def __init__(self, parent, name, dictinputs=None):
            CommonOptimetrics.__init__(self, parent, name, dictinputs=dictinputs, optimtype="OptiOptimization")
            pass

        @aedt_exception_handler
        def add_goal(self, calculation="", calculation_value="", calculation_type="discrete", calculation_stop="",
                     reporttype="Modal Solution Data", solution=None, domain="Sweep", goal_name=None, goal_value=1,
                     goal_weight=1, condition="=="):
            """Add a calculation to the analysis.

            Parameters
            ----------
            calculation : str, optional
                Expression for the calculation, such as ``"dB(S(1,1,))"``. The default is ``""``.
            calculation_value : str, optional
                Value for the calculation, such as ``"1GHz"`` if a sweep. The default is ``""``.
                If ``calculation_type="discrete"``, the value is discrete or is a list. If the
                value is a range, it is the starting value. 
            calculation_type : str, optional
                Type of the calculation. Options are ``"discrete"`` or ``"range"``. 
                The default is ``"discrete"``.
            calculation_stop : str, optional
                Stopping value if ``calculation_type="range"``.  The default is ``""``.
            reporttype : str, optional
                Name of the report to add the calculation to. The default 
                is ``"Modal Solution Data"``.
            solution : str, optional
                Type of the solution. The default is ``None``, in which case the default 
                solution is used. 
            domain : str, optional
                Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
            goal_name : str, optional
                 Name of the goal. The default is ``None``.
            goal_value : optional
                 Value for the goal. The default is ``1``.
            goal_weight : optional
                 Value for the goal weight. The default is ``1``.
            condition : string, optional
                 The default is ``"=="``.

            Returns
            -------
            bool
                ``True`` when successful, ``False`` when failed.
            
            """
            return self._add_goal(optigoalname="Goals", reporttype=reporttype, solution=solution, domain=domain,
                                  calculation_type=calculation_type, calculation=calculation,
                                  calc_val1=calculation_value, calc_val2=calculation_stop, goal_name=goal_name, goal_value=goal_value, goal_weight=goal_weight, condition=condition)

    @property
    def parent(self):
        """Parent object."""
        return self._parent

    @property
    def optimodule(self):
        """Optimetrics module.
        
        Returns
        -------
        :class:`Optimetrics`
            Optimetrics object.
        
        """
        return self.parent.odesign.GetModule("Optimetrics")

    def __init__(self, parent):
        self._parent = parent
        self.setups = []
        if self._parent.design_properties:
            try:
                setups_data = self._parent.design_properties["Optimetrics"]["OptimetricsSetups"]
                for data in setups_data:
                    if type(setups_data[data]) is OrderedDict and setups_data[data]['SetupType'] == "OptiOptimization":
                        self.setups.append(self.Setup(parent, data, setups_data[data]))
            except:
                pass



    @aedt_exception_handler
    def add_optimization(self, calculation, calculation_value, calculation_type="Freq",
                         reporttype="Modal Solution Data", domain="Sweep", condition="<=", goal_value=1, goal_weight=1,
                         solution=None, parametricname=None):
        """Add a basic optimization analysis. 
        
        You can customize all options after the analysis is added. 

        Parameters
        ----------
        calculation : str, optional
            Name of the calculation.
        calculation_value : str, optional
            Variation value, such as ``"1GHz"``.
        calculation_type : str, optional
            Type of the calculation. The default is ``"Freq"``.
        reporttype : str, optional
            Name of the report to add the calculation to. The default 
            is ``"Modal Solution Data"``.
        domain : str, optional
            Type of the domain. The default is ``"Sweep"``. If ``None``, one sweep is taken.
        condition : string, optional
            The default is ``"<="``.
        goal_value : optional
            Value for the goal. The default is ``1``.
        goal_weight : optional
            Value for the goal weight. The default is ``1``.
        solution : str, optional
            Type of the solution. The default is ``None``, in which case the default 
            solution is used.
        parametricname : str, optional
            Name of the analysis. The default is ``None``, in which case a
            default name is assigned.
        
        Returns
        -------
        type
            Optimization object.

        """
        if not parametricname:
            parametricname = generate_unique_name("Optimization")
        setup = self.Setup(self._parent, parametricname)
        sweepdefinition = OrderedDict()
        sweepdefinition["ReportType"] = reporttype
        if not solution:
            solution = self._parent.nominal_sweep
        sweepdefinition["Solution"] = solution
        sweepdefinition["SimValueContext"] = OrderedDict({"Domain": domain})
        sweepdefinition["Calculation"] = calculation
        sweepdefinition["Name"] = calculation
        sweepdefinition["Ranges"] = OrderedDict({"Range": ["Var:=", calculation_type,	"Type:=", "d", "DiscreteValues:=", calculation_value]})
        sweepdefinition["Condition"] = condition
        sweepdefinition["GoalValue"] = OrderedDict(
            {"GoalValueType": "Independent", "Format": "Real/Imag", "bG": ["v:=", "[{};]".format(goal_value)]})
        sweepdefinition["Weight"] = "[{};]".format(goal_weight)

        setup.props["Goals"]["Goal"] = sweepdefinition
        setup.create()
        self.setups.append(setup)
        return setup



"""
This module initializes Siwave and MessageManager and manages Siwave release and closing.

The ``siwave`` module can be initialized as standalone before launching an app or automatically initialized by an app to the latest installed AEDT version.


================

"""
from __future__ import absolute_import
from .generic.general_methods import aedt_exception_handler, generate_unique_name
import os
import sys
import pkgutil

from .misc import list_installed_ansysem

_pythonver = sys.version_info[0]

if "IronPython" in sys.version or ".NETFramework" in sys.version:
    import clr  # IronPython C:\Program Files\AnsysEM\AnsysEM19.4\Win64\common\IronPython\ipy64.exe
    _com = 'pythonnet'
    import System
elif os.name == 'nt':
    modules = [tup[1] for tup in pkgutil.iter_modules()]
    if 'clr' in modules:
        import clr
        import win32com.client
        _com = 'pythonnet_v3'
    elif 'win32com' in modules:
        import win32com.client
        _com = 'pywin32'
    else:
        raise Exception("Error. No win32com.client or Pythonnet modules found. Please install them")


# if _pythonver == 3:
#     from .MessageManager import AEDTMessageManager
# else:
#     from MessageManager import AEDTMessageManager


class Siwave:
    """
    This class initializes Siwave based on the inputs provided.

    Parameters
    ----------

    Returns
    -------

    """
    @property
    def version_keys(self):
        """ """

        self._version_keys = []
        self._version_ids = {}

        version_list = list_installed_ansysem()
        for version_env_var in version_list:
            current_version_id = version_env_var.replace("ANSYSEM_ROOT", '')
            version = int(current_version_id[0:2])
            release = int(current_version_id[2])
            if version < 20:
                if release < 3:
                    version -= 1
                else:
                    release -= 2
            v_key = "20{0}.{1}".format(version, release)
            self._version_keys.append(v_key)
            self._version_ids[v_key] = version_env_var
        return self._version_keys

    @property
    def current_version(self):
        """ """
        return self.version_keys[0]

    def __init__(self, specified_version=None):
        self._main = sys.modules['__main__']
        print ("Launching Siwave Init")
        if  "oSiwave" in dir(self._main) and self._main.oSiwave is not None:
            self._main.AEDTVersion = self._main.oSiwave.GetVersion()[0:6]
            self._main.oSiwave.RestoreWindow()
            #self._main.oMessenger = AEDTMessageManager()
            specified_version = self.current_version
            assert specified_version in self.version_keys, \
                "Specified version {} not known.".format(specified_version)
            version_key = specified_version
            base_path = os.getenv(self._version_ids[specified_version])
            self._main.sDesktopinstallDirectory = base_path
        else:
            if specified_version:
                assert specified_version in self.version_keys, \
                    "Specified version {} not known.".format(specified_version)
                version_key = specified_version
            else:
                version_key = self.current_version
            base_path = os.getenv(self._version_ids[version_key])
            self._main = sys.modules['__main__']
            self._main.sDesktopinstallDirectory = base_path
            version = "Siwave.Application." + version_key
            self._main.AEDTVersion = version_key
            self._main.interpreter = _com
            self._main.interpreter_ver = _pythonver
            if "oSiwave" in dir(self._main):
                del self._main.oSiwave

            if _com == 'pythonnet':
                self._main.oSiwave = System.Activator.CreateInstance(System.Type.GetTypeFromProgID(version))

            elif _com == 'pythonnet_v3': #TODO check if possible to use pythonnet. at the moment the tool open AEDt but doesn't return the wrapper of oApp
                print("Launching AEDT with Module win32com")

                self._main.oSiwave=win32com.client.Dispatch("Siwave.Application.2021.1")

            self._main.AEDTVersion = version_key
            self.oSiwave = self._main.oSiwave
            self._main.oSiwave.RestoreWindow()
            #self._main.oMessenger = AEDTMessageManager()
        self._main.siwave_initialized = True
        self._oproject = self.oSiwave.GetActiveProject()
        pass
        # self.logger = logging.getLogger(__name__)
        # if not self.logger.handlers:
        #     project_dir = self._main.oSiwave.GetProjectDirectory()
        #     logging.basicConfig(
        #         filename=os.path.join(project_dir, "pyaedt.log"),
        #         level=logging.DEBUG,
        #         format='%(asctime)s:%(name)s:%(levelname)-8s:%(message)s',
        #         datefmt='%Y/%m/%d %H.%M.%S')
        #     self.logger = logging.getLogger(__name__)

        # info_msg1 = 'Started external COM connection with module {0}'.format(_com)
        # info_msg2 = 'Python version {0}'.format(sys.version)
        # info_msg3 = 'Exe path: {0}'.format(sys.executable)
        # self._main.oMessenger.add_info_message(info_msg1, 'Global')
        # self._main.oMessenger.add_info_messge(info_msg2, 'Global')
        # self._main.oMessenger.add_info_message(info_msg3, 'Global')



    @property
    def project_name(self):
        """ """
        return self._oproject.GetName()


    @property
    def project_path(self):
        """ """
        return os.path.normpath(self.oSiwave.GetProjectDirectory())

    @property
    def project_file(self):
        """ """
        return os.path.join(self.project_path, self.project_name + '.siw')

    @property
    def lock_file(self):
        """ """
        return os.path.join(self.project_path, self.project_name + '.siw.lock')

    @property
    def results_directory(self):
        """ """
        return os.path.join(self.project_path, self.project_name + '.siwresults')


    @property
    def src_dir(self):
        """ """
        return os.path.dirname(os.path.realpath(__file__))

    @property
    def pyaedt_dir(self):
        """ """
        return os.path.realpath(os.path.join(self.src_dir, '..'))


    @property
    def oproject(self):
        """ """
        return self._oproject

    @aedt_exception_handler
    def open_project(self, proj_path=None):
        """

        Parameters
        ----------
        proj_path : str, optional
            Full path to the project. The default is ``None``.

        Returns
        -------

        """

        if os.path.exists(proj_path):
            self.oSiwave.OpenProject(proj_path)
            self._oproject = self.oSiwave.GetActiveProject()


    @aedt_exception_handler
    def save_project(self, projectpath=None, projectName=None):
        """

        Parameters
        ----------
        proj_path : str, optional
            Full path to the project. The default is ``None``.
        projectName : str, optional
             Name of the project. The default is ``None``.

        Returns
        -------

        """
        if projectName and projectpath:
            self.oproject.ScrSaveProjectAs(os.path.join(projectpath, projectName+".siw"))
        else:
            self.oproject.Save()
        return True


    @aedt_exception_handler
    def quit_application(self):
        """ """

        self._main.oSiwave.Quit()
        return True



import random
from collections import defaultdict

from ..generic.general_methods import aedt_exception_handler, retry_ntimes
from .Object3d import CircuitComponent

class CircuitComponents(object):
    """CircutComponents class. 
    
     This is the common class for managing all circuit components for Nexxim and Simplorer.
     """

    @property
    def oeditor(self):
        """Editor object."""
        return self.modeler.oeditor

    @property
    def _messenger(self):
        """_messenger."""
        return self._parent._messenger

    @property
    def version(self):
        """Version."""
        return self._parent._aedt_version

    @property
    def design_types(self):
        """Design types."""
        return self._parent._modeler

    @property
    def model_units(self):
        """Model units."""
        return self.modeler.model_units

    @property
    def o_model_manager(self):
        """Model manager object."""
        return self.modeler.o_model_manager

    @property
    def o_definition_manager(self):
        """Definition manager object."""
        return self._parent._oproject.GetDefinitionManager()

    @property
    def o_symbol_manager(self):
        """Symbol manager object."""
        return self.o_definition_manager.GetManager("Symbol")

    @property
    def o_component_manager(self):
        """Component manager object."""
        return self.o_definition_manager.GetManager("Component")

    @property
    def design_type(self):
        """Design type."""
        return self._parent.design_type

    @aedt_exception_handler
    def __getitem__(self, partname):
        """Retrieve a part.

        Parameters
        ----------
        partname : int or str
           Part ID or part name. 
        
        Returns
        -------
        type
            Part object details.
        """
        if type(partname) is int:
            return self.components[partname]
        for el in self.components:
            if self.components[el].name == partname or self.components[el].composed_name == partname or el == partname:
                return self.components[el]

        return None

    def __init__(self, parent, modeler):
        self._parent = parent
        self.modeler = modeler
        self._currentId = 0
        self.components = defaultdict(CircuitComponent)
        pass

    @aedt_exception_handler
    def create_unique_id(self):
        """Create an unique ID.
        
        Returns
        -------
        int
            Unique ID in the range of ``[1, 65535]``.
            
        """
        id = random.randint(1, 65535)
        while id in self.components:
            id = random.randint(1, 65535)
        return id

    @aedt_exception_handler
    def create_wire(self, points_array):
        """Create a wire.
        Parameters
        ----------
        points_array : list
            A nested list of point coordinates. For example, ``[[x1, y1], [x2, y2]....]``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        pointlist = [str(tuple(i)) for i in points_array]
        self.oeditor.CreateWire(
            [
                "NAME:WireData",
                "Name:=", "",
                "Id:=", random.randint(20000, 23000),
                "Points:=", pointlist
            ],
            [
                "NAME:Attributes",
                "Page:=", 1
            ])
        return True

    @aedt_exception_handler
    def create_iport(self, name, posx=0.1, posy=0.1, angle=0):
        """Create a port.

        Parameters
        ----------
        name : str
            Name of the port.
        posx : float, optional
            Position on the X axis. The default is ``0.1``.
        posy : float, optional
            Position on the Y axis. The default is ``0.1``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.

        Returns
        -------
        type
            Port object.
        str
           Port name.

        """
        id = self.create_unique_id()
        arg1 = ["NAME:IPortProps", "Name:=", name, "Id:=", id]
        arg2 = ["NAME:Attributes", "Page:=", 1, "X:=", posx, "Y:=", posy, "Angle:=", angle, "Flip:=", False]
        id = self.oeditor.CreateIPort(arg1, arg2)

        id = int(id.split(";")[1])
        self.add_id_to_component(id)
        # return id, self.components[id].composed_name
        for el in self.components:
            if name in self.components[el].composed_name:
                return el, self.components[el].composed_name

    @aedt_exception_handler
    def create_page_port(self, name, posx=0.1, posy=0.1, angle=0):
        """Create a page port.

        Parameters
        ----------
        name : str
            Name of the port.
        posx : float, optional
            Position on the X axis. The default is ``0.1``.
        posy : float, optional
            Position on the Y axis. The default is ``0.1``.
        angle : optional
            Angle rotation in degrees. The default is ``0``.

        Returns
        -------
        type
            Port ID.
        str
            Port name.

        """
        id = self.create_unique_id()
        id = self.oeditor.CreatePagePort(
            [
                "NAME:PagePortProps",
                "Name:=", name,
                "Id:=", id
            ],
            [
                "NAME:Attributes",
                "Page:=", 1,
                "X:=", posx,
                "Y:=", posy,
                "Angle:=", angle,
                "Flip:=", False
            ])
        id = int(id.split(";")[1])
        # self.refresh_all_ids()
        self.add_id_to_component(id)
        return id, self.components[id].composed_name

    @aedt_exception_handler
    def create_gnd(self, posx, posy):
        """Create a ground.

        Parameters
        ----------
        posx : float, optional
            Position on the X axis. The default is ``0.1``.
        posy : float, optional
            Position on the Y axis. The default is ``0.1``.
       
        Returns
        -------
        type
            Ground object.
        str
            Ground name.

        """
        id = self.create_unique_id()

        name = self.oeditor.CreateGround(
            [
                "NAME:GroundProps",
                "Id:=", id
            ],
            [
                "NAME:Attributes",
                "Page:=", 1,
                "X:=", posx,
                "Y:=", posy,
                "Angle:=", 0,
                "Flip:=", False
            ])
        id = int(name.split(";")[1])
        self.add_id_to_component(id)
        # return id, self.components[id].composed_name
        for el in self.components:
            if name in self.components[el].composed_name:
                return el, self.components[el].composed_name

    @aedt_exception_handler
    def create_model_from_touchstone(self, touchstone_full_path, model_name=None):
        """Create a model from a Touchstone file.

        Parameters
        ----------
        touchstone_full_path : str
            Full path to the Touchstone file.
        model_name : str, optional
            Name of the model. The default is ``None``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.     

        """
        if not model_name:
            model_name = os.path.splitext(os.path.basename(touchstone_full_path))[0]

        num_terminal = int(touchstone_full_path[-2:-1])
        arg = ["NAME:" + model_name, "Name:=", model_name, "ModTime:=", 0, "Library:=", "", "LibLocation:=", "Project",
               "ModelType:=", "nport", "Description:=", "", "ImageFile:=", "", "SymbolPinConfiguration:=", 0,
               ["NAME:PortInfoBlk"], ["NAME:PortOrderBlk"], "filename:=", touchstone_full_path, "numberofports:=",
               num_terminal, "sssfilename:=", "", "sssmodel:=", False, "PortNames:=",
               ["Port" + str(i + 1) for i in range(num_terminal)], "domain:=", "frequency", "datamode:=",
               "Link", "devicename:=", "", "SolutionName:=", "", "displayformat:=", "MagnitudePhase", "datatype:=",
               "SMatrix", ["NAME:DesignerCustomization", "DCOption:=", 0, "InterpOption:=", 0, "ExtrapOption:=", 1,
                           "Convolution:=", 0, "Passivity:=", 0, "Reciprocal:=", False, "ModelOption:=", "",
                           "DataType:=", 1],
               ["NAME:NexximCustomization", "DCOption:=", 3, "InterpOption:=", 1, "ExtrapOption:=", 3, "Convolution:=",
                0, "Passivity:=", 0, "Reciprocal:=", False, "ModelOption:=", "", "DataType:=", 2],
               ["NAME:HSpiceCustomization", "DCOption:=", 1, "InterpOption:=", 2, "ExtrapOption:=", 3, "Convolution:=",
                0, "Passivity:=", 0, "Reciprocal:=", False, "ModelOption:=", "", "DataType:=", 3], "NoiseModelOption:=",
               "External"]
        self.o_model_manager.Add(arg)
        arg = ["NAME:" + model_name, "Info:=",
               ["Type:=", 10, "NumTerminals:=", num_terminal, "DataSource:=", "", "ModifiedOn:=", 1618569625,
                "Manufacturer:=",
                "", "Symbol:=", "", "ModelNames:=", "", "Footprint:=", "", "Description:=", "", "InfoTopic:=", "",
                "InfoHelpFile:=", "", "IconFile:=", "nport.bmp", "Library:=", "", "OriginalLocation:=", "Project",
                "IEEE:=", "", "Author:=", "", "OriginalAuthor:=", "", "CreationDate:=", 1618569625, "ExampleFile:=",
                "", "HiddenComponent:=", 0, "CircuitEnv:=", 0, "GroupID:=", 0],
               "CircuitEnv:=", 0, "Refbase:=", "S", "NumParts:=", 1,
               "ModSinceLib:=", False]
        for i in range(num_terminal):
            arg.append("Terminal:=")
            arg.append(["Port" + str(i + 1), "Port" + str(i + 1), "A", False, i + 6, 1, "", "Electrical", "0"])
        arg.append("CompExtID:=")
        arg.append(num_terminal)
        arg.append(["NAME:Parameters", "MenuProp:=", ["CoSimulator", "SD", "", "Default", 0], "ButtonProp:=",
                    ["CosimDefinition", "SD", "", "Edit", "Edit", 40501, "ButtonPropClientData:=", []]])
        arg.append(["NAME:CosimDefinitions",
                    ["NAME:CosimDefinition", "CosimulatorType:=", 102, "CosimDefName:=", "Default", "IsDefinition:=",
                     True, "Connect:=", True, "ModelDefinitionName:=", model_name, "ShowRefPin2:=", 2, "LenPropName:=",
                     ""], "DefaultCosim:=", "Default"])

        self.o_component_manager.Add(arg)

    @aedt_exception_handler
    def create_component_from_touchstonmodel(self, modelname,xpos=0.1, ypos=0.1, angle=0,):
        """Create a component from a Touchstone model.

        Parameters
        ----------
        modelname : str
            Name of the Touchstone model.
        xpos : float, optional
            Position on the X axis. The default is ``0.1``.
        ypos : float, optional
            Position on the Y axis. The default is ``0.1``.
        angle : float, optional
            Angle rotation in degrees. The default is ``0``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        id = self.create_unique_id()
        id = self.oeditor.CreateComponent(["NAME:ComponentProps", "Name:=", modelname, "Id:=", str(id)],
                                ["NAME:Attributes", "Page:=", 1, "X:=", xpos, "Y:=", ypos, "Angle:=", angle, "Flip:=",
                                 False])
        id = int(id.split(";")[1])
        self.add_id_to_component(id)
        return id, self.components[id].composed_name


    @aedt_exception_handler
    def create_component(self, inst_name=None, component_library="Resistors",
                         component_name="RES_", xpos=0.1, ypos=0.1, angle=0, use_instance_id_netlist=False,
                         global_netlist_list=[]):
        """Create a component from a library.

        Parameters
        ----------
        inst_name : str, optional
            Name of the instance. The default is ``None.``
        component_library : str, optional
            Name of the component library. The default is ``"Resistors"``.
        component_name : str, optional
            Name of component in the library. The default is ``"RES"``.
        xpos : float, optional
            Position on the X axis. The default is ``0.1``.
        yos : float, optional
            Position on the Y axis. The default is ``0.1``.
        angle : optional
            Angle rotation in degrees. The default is ``0``.
        use_instance_id_netlist : bool, optional
            Whether to enable the instance ID in the net list.
            The default is ``False``.
        global_netlist_list : list, optional
            The default is``[]``.

        Returns
        -------
        type
            Component ID.
        str
            Component name.

        """
        id = self.create_unique_id()
        if component_library:
            name = self.design_libray + "\\" + component_library + ":" + component_name
        else:
            name = component_name
        id = self.oeditor.CreateComponent(
            [
                "NAME:ComponentProps",
                "Name:=", name,
                "Id:=", str(id)
            ],
            [
                "NAME:Attributes",
                "Page:=", 1,
                "X:=", xpos,
                "Y:=", ypos,
                "Angle:=", angle,
                "Flip:=", False
            ])
        id = int(id.split(";")[1])
        # self.refresh_all_ids()
        self.add_id_to_component(id)
        if inst_name:
            self.components[id].set_property("InstanceName", inst_name)
        if use_instance_id_netlist:
            self.enable_use_instance_name(component_library, component_name)
        elif global_netlist_list:
            self.enable_global_netlist(component_name, global_netlist_list)
        return id, self.components[id].composed_name

    @aedt_exception_handler
    def disable_data_netlist(self, component_name):
        """Disable the Nexxim global net list.

        Parameters
        ----------
        component_name : str
            Name of the component.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        name = component_name

        properties = self.o_component_manager.GetData(name)
        if len(properties) > 0:
            nexxim = list(properties[len(properties) - 1][1])
            for el in nexxim:
                if el == "Data:=":
                    nexxim_data = list(nexxim[nexxim.index(el) + 1])
                    nexxim_data[1] = ""
                    nexxim[nexxim.index(el) + 1] = nexxim_data
        self.o_component_manager.Edit(name, ["Name:" + component_name,
                                             ["NAME:CosimDefinitions", nexxim, "DefaultCosim:=", "DefaultNetlist"]])
        return True

    @aedt_exception_handler
    def enable_global_netlist(self, component_name, global_netlist_list=[]):
        """Enable Nexxim global net list.

        Parameters
        ----------
        component_name : str
            Name of the component.
        global_netlist_list : list
            A list of lines to include. The default is ``[]``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        name = component_name

        properties = self.o_component_manager.GetData(name)
        if len(properties) > 0:
            nexxim = list(properties[len(properties) - 1][1])
            for el in nexxim:
                if el == "GRef:=":
                    nexxim_data = list(nexxim[nexxim.index(el) + 1])
                    nexxim_data[1] = "\n".join(global_netlist_list).replace("\\", "/")
                    nexxim[nexxim.index(el) + 1] = nexxim_data
        retry_ntimes(10, self.o_component_manager.Edit, name, ["Name:" + component_name,
                                             ["NAME:CosimDefinitions", nexxim, "DefaultCosim:=", "DefaultNetlist"]])
        return True

    @aedt_exception_handler
    def create_symbol(self, symbol_name, pin_lists):
        """Create a symbol.

        Parameters
        ----------
        symbol_name : str
            Name of the symbol.
        pin_lists : list
            List of the pins.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """

        numpins = len(pin_lists)
        h = int(numpins / 2)
        x1 = 0
        y2 = 0
        x2 = 0.00508
        y1 = 0.00254 * (h + 3)
        xp = -0.00254
        yp = 0.00254 * (h + 2)
        angle = 0
        self.o_symbol_manager.Add(
            ["NAME:" + symbol_name, "ModTime:=", 1591858230, "Library:=", "", "ModSinceLib:=", False, "LibLocation:=",
             "Project", "HighestLevel:=", 1, "Normalize:=", True, "InitialLevels:=", [0, 1], ["NAME:Graphics"]])
        arg = ["NAME:" + symbol_name, "ModTime:=", 1591858265, "Library:=", "", "ModSinceLib:=", False, "LibLocation:=",
               "Project", "HighestLevel:=", 1, "Normalize:=", False, "InitialLevels:=", [0, 1]]
        oDefinitionEditor = self._parent._oproject.SetActiveDefinitionEditor("SymbolEditor", symbol_name)
        id = 2
        oDefinitionEditor.CreateRectangle(["NAME:RectData", "X1:=", x1, "Y1:=", y1, "X2:=", x2, "Y2:=", y2,
                                           "LineWidth:=", 0, "BorderColor:=", 0, "Fill:=", 0, "Color:=", 0, "Id:=", id],
                                          ["NAME:Attributes", "Page:=", 1])
        i = 1
        id += 2
        r = numpins - (h * 2)
        for pin in pin_lists:
            oDefinitionEditor.CreatePin(["NAME:PinData", "Name:=", pin, "Id:=", id],
                                        ["NAME:PinParams", "X:=", xp, "Y:=", yp, "Angle:=", angle, "Flip:=", False])
            arg.append(["NAME:PinDef", "Pin:=",
                        [pin, xp, yp, angle, "N", 0, 0.00254, False, 0, True, "", False, False, pin,
                         True]])
            if i == (h + r):
                yp = 0.00254 * (h + 2)
                xp = 0.00762
                angle = 3.14159265358979
            else:
                yp -= 0.00254
            id += 2
            i += 1

        arg.append(["NAME:Graphics",
                    ["NAME:1", "Rect:=", [0, 0, 0, 0, (x1 + x2) / 2, (y1 + y2) / 2, x2 - x1, y1 - y2, 0, 0, 8192]]])
        self.o_symbol_manager.EditWithComps(symbol_name, arg, [])
        oDefinitionEditor.CloseEditor()
        return True

    @aedt_exception_handler
    def create_new_component_from_symbol(self, symbol_name, pin_lists, Refbase="U", parameter_list=[],
                                         parameter_value=[]):
        """Create a component from a symbol.

        Parameters
        ----------
        symbol_name : str
            Name of the symbol.
        pin_lists : list
            List of the pins.
        Refbase : str, optional
            Reference base. The default is ``"U"``.
        parameter_list : list, optional
            List of the parameters. The default is ``[]``.
        parameter_value : list, optional
            List of the parameter values. The default is ``[]``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        arg = ["NAME:" + symbol_name, "Info:=",
               ["Type:=", 0, "NumTerminals:=", 5, "DataSource:=", "", "ModifiedOn:=", 1591858313, "Manufacturer:=", "",
                "Symbol:=", symbol_name, "ModelNames:=", "", "Footprint:=", "",
                "Description:=", "", "InfoTopic:=", "", "InfoHelpFile:=", "", "IconFile:=", "", "Library:=", "",
                "OriginalLocation:=", "Project", "IEEE:=", "", "Author:=", "", "OriginalAuthor:=", "",
                "CreationDate:=", 1591858278, "ExampleFile:=", "", "HiddenComponent:=", 0, "CircuitEnv:=", 0,
                "GroupID:=", 0], "CircuitEnv:=", 0, "Refbase:=", Refbase, "NumParts:=", 1, "ModSinceLib:=", True]

        for pin in pin_lists:
            arg.append("Terminal:=")
            arg.append([pin, pin, "A", False, 0, 1, "", "Electrical", "0"])
        arg.append("CompExtID:=")
        arg.append(1)
        arg2 = ["NAME:Parameters"]
        for el, val in zip(parameter_list, parameter_value):
            if type(val) is str:
                arg2.append("TextValueProp:=")
                arg2.append([el, "D", "", val])
            else:
                arg2.append("ValueProp:=")
                arg2.append([el, "D", "", val, False, ""])
        arg2.append("ButtonProp:=")
        arg2.append(["CosimDefinition", "D", "", "Edit", "Edit", 40501, "ButtonPropClientData:=", []])
        arg2.append("MenuProp:=")
        arg2.append(["CoSimulator", "D", "", "DefaultNetlist", 0])

        arg.append(arg2)
        spicesintax = Refbase + "@ID "
        id = 0
        while id < (len(pin_lists) - 1):
            spicesintax += "%" + str(id) + " "
            id += 1
        for el in parameter_list:
            spicesintax += "@{} ".format(el)

        arg3 = [
            "NAME:CosimDefinitions",
            [
                "NAME:CosimDefinition",
                "CosimulatorType:=", 4,
                "CosimDefName:=", "DefaultNetlist",
                "IsDefinition:=", True,
                "Connect:=", True,
                "Data:=", ["Nexxim Circuit:=", spicesintax],
                "GRef:=", ["Nexxim Circuit:=", ""]
            ],
            "DefaultCosim:=", "DefaultNetlist"
        ]
        arg.append(arg3)
        self.o_component_manager.Add(arg)
        return True

    @aedt_exception_handler
    def enable_use_instance_name(self, component_library="Resistors",
                                 component_name="RES_"):
        """

        Parameters
        ----------
        component_library : str, optional
             Name of the component library. The default is ``"Resistors"``.
        component_name : str, optional
             Name of the component. The default is ``"RES_"``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        if component_library:
            name = self.design_libray + "\\" + component_library + ":" + component_name
        else:
            name = component_name

        properties = self.o_component_manager.GetData(name)
        if len(properties) > 0:
            nexxim = list(properties[len(properties) - 1][1])
            for el in nexxim:
                if el == "Data:=":
                    nexxim_data = list(nexxim[nexxim.index(el) + 1])
                    netlist = nexxim_data[1]
                    if "@InstanceName" not in netlist[:15]:
                        newnetlist = "@InstanceName" + netlist[4:]
                        nexxim_data[1] = newnetlist
                    nexxim[nexxim.index(el) + 1] = nexxim_data
                elif el == "GRef:=":
                    nexxim_data = list(nexxim[nexxim.index(el) + 1])
                    nexxim[nexxim.index(el) + 1] = nexxim_data
        retry_ntimes(10, self.o_component_manager.Edit, name, ["Name:" + component_name,
                                             ["NAME:CosimDefinitions", nexxim, "DefaultCosim:=", "DefaultNetlist"]])
        return True

    @aedt_exception_handler
    def refresh_all_ids(self):
        """Refresh all IDs and return the number of components."""
        obj = self.oeditor.GetAllComponents()
        for el in obj:
            if not self.get_obj_id(el):
                name = el.split(";")
                o = CircuitComponent(self.oeditor, tabname=self.tab_name)
                o.name = name[0]
                o.id = int(name[1])
                o.schematic_id = name[2]
                o_update = self.update_object_properties(o)
                objID = o.id
                self.components[objID] = o_update
        return len(self.components)

    @aedt_exception_handler
    def add_id_to_component(self, id):
        """Add an ID to a component.

        Parameters
        ----------
        id : int
            ID to assign the component.

        Returns
        -------
        int
            Number of components.

        """
        obj = retry_ntimes(10, self.oeditor.GetAllElements)
        for el in obj:
            name = el.split(";")
            if len(name) > 1 and str(id) == name[1]:
                o = CircuitComponent(self.oeditor, tabname=self.tab_name)
                o.name = name[0]
                if len(name) > 2:
                    o.id = int(name[1])
                    o.schematic_id = int(name[2])
                    objID = o.id
                else:
                    o.schematic_id = int(name[1])
                    objID = o.schematic_id
                o_update = self.update_object_properties(o)

                self.components[objID] = o_update
        return len(self.components)

    @aedt_exception_handler
    def get_obj_id(self, objname):
        """Retrieve the ID of an object.

        Parameters
        ----------
        objname : str
            Name of the object.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        for el in self.components:
            if self.components[el].name == objname:
                return el
        return None

    @aedt_exception_handler
    def update_object_properties(self, o):
        """Update the properties of an object.

        Parameters
        ----------
        o :
            Object to update.

        Returns
        -------
        type
            Object with properties.

        """
        name = o.composed_name
        proparray = retry_ntimes(10, self.oeditor.GetProperties, "PassedParameterTab", name)
        for j in proparray:
            propval = retry_ntimes(10, self.oeditor.GetPropertyValue, "PassedParameterTab", name, j)
            o._add_property(j, propval)
        return o

    @aedt_exception_handler
    def get_pins(self, partid):
        """Retrieve one or more pins.

        Parameters
        ----------
        partid : int or str
            One or more IDs or names for the pins to retrieve.

        Returns
        -------
        type
            Pin with properties.

        """
        if type(partid) is str:
            pins = retry_ntimes(10, self.oeditor.GetComponentPins, partid)
            #pins = self.oeditor.GetComponentPins(partid)
        else:
            pins = retry_ntimes(10, self.oeditor.GetComponentPins, self.components[partid].composed_name)
            #pins = self.oeditor.GetComponentPins(self.components[partid].composed_name)
        return list(pins)

    @aedt_exception_handler
    def get_pin_location(self, partid, pinname):
        """Retrieve the location of a pin.

        Parameters
        ----------
        partid : int
            ID of the part.
        pinname :
            Name of the pin.

        Returns
        -------
        List
            List of axis values ``[x, y]``.

        """
        if type(partid) is str:
            x = retry_ntimes(30, self.oeditor.GetComponentPinLocation, partid, pinname, True)
            y = retry_ntimes(30, self.oeditor.GetComponentPinLocation, partid, pinname, False)
        else:
            x = retry_ntimes(30, self.oeditor.GetComponentPinLocation, self.components[partid].composed_name, pinname, True)
            y = retry_ntimes(30, self.oeditor.GetComponentPinLocation, self.components[partid].composed_name, pinname, False)
        return [x, y]

    @aedt_exception_handler
    def arg_with_dim(self, Value, sUnits=None):
        """

        Parameters
        ----------
        Value : str
            
        sUnits :
            The default is ``None``.

        Returns
        -------
        type
            

        """
        if type(Value) is str:
            val = Value
        else:
            if sUnits is None:
                sUnits = self.model_units
            val = "{0}{1}".format(Value, sUnits)

        return val



"""
This module contains the ``PostProcessor`` class.

It contains all advanced postprocessing functionalities that require Python 3.x packages like NumPy and Matplotlib.
"""
from __future__ import absolute_import
import os

from .PostProcessor import PostProcessor as Post
from ..generic.general_methods import aedt_exception_handler
import time
import math
import warnings

try:
    import numpy as np
except ImportError:
    warnings.warn("The NumPy module is required to run some functionalities of PostProcess.\n"
                  "Install with \n\npip install numpy\n\nRequires CPython")

try:
    import pyvista as pv
    pyvista_available = True
except ImportError:
    warnings.warn("The PyVista module required to run some functionalities of PostProcess.\n"
                  "Install with \n\npip install pyvista\n\nRequires CPython")

try:
    from IPython.display import Image, display
    ipython_available = True
except ImportError:
    warnings.warn("The Ipython module required to run some functionalities of PostProcess.\n"
                  "Install with \n\npip install ipython\n\nRequires CPython")

try:
    import matplotlib.pyplot as plt
except ImportError:
    warnings.warn("The Matplotlib module required to run some functionalities of PostProcess.\n"
                  "Install with \n\npip install matplotlib\n\nRequires CPython")


def is_float(istring):
    """Convert a string to a float.

    Parameters
    ----------
    istring : str
        String to convert to a float.

    Returns
    -------
    float
        Converted float when successful, ``0`` when when failed. 
    """
    try:
        return float(istring.strip())
    except Exception:
        return 0

class PostProcessor(Post):
    """ """
    def __init__(self, parent):
        Post.__init__(self, parent)

    @aedt_exception_handler
    def nb_display(self, show_axis=True, show_grid=True, show_ruler=True):
        """Jupyter Notebook is not supported by IronPython!"""
        file_name = self.export_model_picture(show_axis=show_axis, show_grid=show_grid, show_ruler=show_ruler)
        return Image(file_name, width=500)

    @aedt_exception_handler
    def get_efields_data(self, setup_sweep_name='', ff_setup="Infinite Sphere1", freq='All'):
        """Compute ``Etheta`` and ``EPhi`` and returns an array of ``[theta_range, phi_range, Etheta, Ephi]``.
        
        .. warning::
           This method requires NumPy to be installed on your machine.
        

        Parameters
        ----------
        setup_sweep_name : str, optional
            Name of the setup on which to compute the report. The default is ``""``, which means that
            the nominal adaptive is to be applied.
        ff_setup : str, optional
            Far field setup. The default is ``"Infinite Sphere1"``.
        freq : str, optional
            The default is ``"All"``.

        Returns
        -------
        np.ndarray
            ``numpy`` array containing ``[theta_range, phi_range, Etheta, Ephi]``.
        """
        if not setup_sweep_name:
            setup_sweep_name = self._parent.nominal_adaptive
        results_dict = {}
        all_sources = self.post_osolution.GetAllSources()
        # assuming only 1 mode
        all_sources_with_modes = [s + ':1' for s in all_sources]

        for n, source in enumerate(all_sources_with_modes):
            edit_sources_ctxt = [["IncludePortPostProcessing:=", False, "SpecifySystemPower:=", False]]
            for m, each in enumerate(all_sources_with_modes):
                if n == m:  # set only 1 source to 1W, all the rest to 0
                    mag = 1
                else:
                    mag = 0
                phase = 0
                edit_sources_ctxt.append(
                    ["Name:=", "{}".format(each), "Magnitude:=", "{}W".format(mag), "Phase:=", "{}deg".format(phase)])
            self.post_osolution.EditSources(edit_sources_ctxt)

            ctxt = ['Context:=', ff_setup]

            sweeps = ['Theta:=', ['All'], 'Phi:=', ['All'], 'Freq:=', [freq]]

            trace_name = "rETheta"
            solnData = self.get_far_field_data(setup_sweep_name=setup_sweep_name, domain=ff_setup,
                                               expression=trace_name)

            data = solnData.nominal_variation

            theta_vals = np.degrees(np.array(data.GetSweepValues('Theta')))
            phi_vals = np.degrees(np.array(data.GetSweepValues('Phi')))
            # phi is outer loop
            theta_unique = np.unique(theta_vals)
            phi_unique = np.unique(phi_vals)
            theta_range = np.linspace(np.min(theta_vals), np.max(theta_vals), np.size(theta_unique))
            phi_range = np.linspace(np.min(phi_vals), np.max(phi_vals), np.size(phi_unique))
            real_theta = np.array(data.GetRealDataValues(trace_name))
            imag_theta = np.array(data.GetImagDataValues(trace_name))

            trace_name = "rEPhi"
            solnData = self.get_far_field_data(setup_sweep_name=setup_sweep_name, domain=ff_setup,
                                               expression=trace_name)
            data = solnData.nominal_variation

            real_phi = np.array(data.GetRealDataValues(trace_name))
            imag_phi = np.array(data.GetImagDataValues(trace_name))

            Etheta = np.vectorize(complex)(real_theta, imag_theta)
            Ephi = np.vectorize(complex)(real_phi, imag_phi)
            source_name_without_mode = source.replace(':1', '')
            results_dict[source_name_without_mode] = [theta_range, phi_range, Etheta, Ephi]
        return results_dict

    @aedt_exception_handler
    def ff_sum_with_delta_phase(self, ff_data, xphase=0, yphase=0):
        """

        Parameters
        ----------
        ff_data :
            
        xphase : float, optional
            Phase in x the direction.  The default is ``0``.
        yphase : float, optional
            Phase in the y direction.  The default is ``0``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        array_size = [4, 4]
        loc_offset = 2

        rETheta = ff_data[2]
        rEPhi = ff_data[3]
        weight = np.zeros((array_size[0], array_size[0]))
        mag = np.ones((array_size[0], array_size[0]))
        for m in range(array_size[0]):
            for n in range(array_size[1]):
                mag = mag[m][n]
                ang = np.radians(xphase * m) + np.radians(yphase * n)
                weight[m][n] = np.sqrt(mag) * np.exp(1 * ang)
        return True

    @aedt_exception_handler
    def _triangle_vertex(self, elements_nodes, num_nodes_per_element, take_all_nodes=True):
        """

        Parameters
        ----------
        elements_nodes :
            
        num_nodes_per_element :
            
        take_all_nodes : bool, optional
            The default is ``True``.

        Returns
        -------

        """
        trg_vertex = []
        if num_nodes_per_element == 10 and take_all_nodes:
            for e in elements_nodes:
                trg_vertex.append([e[0], e[1], e[3]])
                trg_vertex.append([e[1], e[2], e[4]])
                trg_vertex.append([e[1], e[4], e[3]])
                trg_vertex.append([e[3], e[4], e[5]])

                trg_vertex.append([e[9], e[6], e[8]])
                trg_vertex.append([e[6], e[0], e[3]])
                trg_vertex.append([e[6], e[3], e[8]])
                trg_vertex.append([e[8], e[3], e[5]])

                trg_vertex.append([e[9], e[7], e[8]])
                trg_vertex.append([e[7], e[2], e[4]])
                trg_vertex.append([e[7], e[4], e[8]])
                trg_vertex.append([e[8], e[4], e[5]])

                trg_vertex.append([e[9], e[7], e[6]])
                trg_vertex.append([e[7], e[2], e[1]])
                trg_vertex.append([e[7], e[1], e[6]])
                trg_vertex.append([e[6], e[1], e[0]])
        elif num_nodes_per_element == 10 and not take_all_nodes:
            for e in elements_nodes:
                trg_vertex.append([e[0], e[2], e[5]])
                trg_vertex.append([e[9], e[0], e[5]])
                trg_vertex.append([e[9], e[2], e[0]])
                trg_vertex.append([e[9], e[2], e[5]])

        elif num_nodes_per_element == 6 and not take_all_nodes:
            for e in elements_nodes:
                trg_vertex.append([e[0], e[2], e[5]])

        elif num_nodes_per_element == 6 and take_all_nodes:
            for e in elements_nodes:
                trg_vertex.append([e[0], e[1], e[3]])
                trg_vertex.append([e[1], e[2], e[4]])
                trg_vertex.append([e[1], e[4], e[3]])
                trg_vertex.append([e[3], e[4], e[5]])

        elif num_nodes_per_element == 4 and take_all_nodes:
            for e in elements_nodes:
                trg_vertex.append([e[0], e[1], e[3]])
                trg_vertex.append([e[1], e[2], e[3]])
                trg_vertex.append([e[0], e[1], e[2]])
                trg_vertex.append([e[0], e[2], e[3]])

        elif num_nodes_per_element == 3:
            trg_vertex = elements_nodes

        return trg_vertex

    @aedt_exception_handler
    def _plot_from_aedtplt(self, aedtplt_files=None, imageformat="jpg", view="iso", plot_type="Full",
                           plot_label="Temperature", model_color="#8faf8f", show_model_edge=False, off_screen=False):
        """Export the 3D field solver mesh, fields, or both mesh and field as images using Python Plotly. 
        
        .. note::
           This function is currently supported only on Windows using CPython.

        Parameters
        ----------
        aedtplt_files : str or list, optional
            Names of the one or more AEDTPLT files generated by AEDT. The default is ``None``.
        imageformat : str, optional
            Format of the image file. Options are ``"jpg"``, ``"png"``, ``"svg"``, and 
            ``"webp"``. The default is ``"jpg"``.
        view : str, optional
            View to export. Options are ``"iso"``, ``"x"`` , ``"y"``, ``"z"``, and ``"all"``. 
            The default is ``"iso"``. The ``"all"`` option exports all views.
        plot_type : str, optional
            Type of the plot. The default is ``"Full"``.
        plot_label : str, optional
            Label for the plot. The default is ``"Temperature"``.
        model_color : str, optional
            Color scheme for the 3D model. The default is ``"#8faf8f"``, which is silver.
        show_model_edge : bool, optional
            Whether to return a list of the files that are generated. The default
            is ``False``.
        off_screen : bool, optional
             The default is ``False``.

        Returns
        -------
        list
            List of exported files.
        """
        start = time.time()
        if type(aedtplt_files) is str:
            aedtplt_files = [aedtplt_files]

        plot = pv.Plotter(off_screen=off_screen)
        if not off_screen:
            plot.enable_anti_aliasing()
        plot.enable_fly_to_right_click()
        lines = []
        for file in aedtplt_files:
            if ".aedtplt" in file:
                with open(file, "r") as f:
                    drawing_found = False
                    for line in f:
                        if "$begin Drawing" in line:
                            drawing_found = True
                            l_tmp = []
                            continue
                        if "$end Drawing" in line:
                            lines.append(l_tmp)
                            drawing_found = False
                            continue
                        if drawing_found:
                            l_tmp.append(line)
                            continue
                        if "Number of drawing:" in line:
                            n_drawings = int(line[18:])
                            continue
            elif ".obj" in file:
                mesh = pv.read(file)

                def create_object_mesh(opacity):
                    """Create the mesh.

                    Parameters
                    ----------
                    opacity :
                        

                    Returns
                    -------

                    """
                    try:
                        plot.remove_actor("Volumes")
                    except:
                        pass
                    plot.add_mesh(mesh, show_scalar_bar=False, opacity=opacity, cmap=[model_color], name="3D Model",
                                  show_edges=show_model_edge, edge_color=model_color)
                plot.add_slider_widget(create_object_mesh, [0, 1], style='modern', value=0.75, pointa=[0.81, 0.98], pointb=[0.95, 0.98], title="Opacity")
        filename = os.path.splitext(aedtplt_files[0])[0]
        print(filename)
        for drawing_lines in lines:
            bounding = []
            elements = []
            nodes_list = []
            solution = []
            for l in drawing_lines:
                if "BoundingBox(" in l:
                    bounding = l[l.find("(") + 1:-2].split(",")
                    bounding = [i.strip() for i in bounding]
                if "Elements(" in l:
                    elements = l[l.find("(") + 1:-2].split(",")
                    elements = [int(i.strip()) for i in elements]
                if "Nodes(" in l:
                    nodes_list = l[l.find("(") + 1:-2].split(",")
                    nodes_list = [float(i.strip()) for i in nodes_list]
                if "ElemSolution(" in l:
                    # convert list of strings to list of floats
                    sols = l[l.find("(") + 1:-2].split(",")
                    sols = [is_float(value) for value in sols]

                    # sols = [float(i.strip()) for i in sols]
                    num_solution_per_element = int(sols[2])
                    sols = sols[3:]
                    sols = [sols[i:i + num_solution_per_element] for i in range(0, len(sols), num_solution_per_element)]
                    solution = [sum(i) / num_solution_per_element for i in sols]

            nodes = [[nodes_list[i], nodes_list[i + 1], nodes_list[i + 2]] for i in range(0, len(nodes_list), 3)]
            num_nodes = elements[0]
            num_elements = elements[1]
            elements = elements[2:]
            element_type = elements[0]
            num_nodes_per_element = elements[4]
            hl = 5  # header length
            elements_nodes = []
            for i in range(0, len(elements), num_nodes_per_element + hl):
                elements_nodes.append([elements[i + hl + n] for n in range(num_nodes_per_element)])
            if solution:
                take_all_nodes = True  # solution case
            else:
                take_all_nodes = False  # mesh case
            trg_vertex = self._triangle_vertex(elements_nodes, num_nodes_per_element, take_all_nodes)
            # remove duplicates
            nodup_list = [list(i) for i in list(set([frozenset(t) for t in trg_vertex]))]
            sols_vertex = []
            if solution:
                sv = {}
                for els, s in zip(elements_nodes, solution):
                    for el in els:
                        if el in sv:
                            sv[el] = (sv[el] + s) / 2
                        else:
                            sv[el] = s
                sols_vertex = [sv[v] for v in sorted(sv.keys())]
            array = [[3] + [j - 1 for j in i] for i in nodup_list]
            faces = np.hstack(array)
            vertices = np.array(nodes)
            surf = pv.PolyData(vertices, faces)
            if sols_vertex:
                temps = np.array(sols_vertex)
                mean = np.mean(temps)
                std = np.std(temps)
                if np.min(temps) > 0:
                    log = True
                else:
                    log = False
                surf.point_arrays[plot_label] = temps

            sargs = dict(title_font_size=10, label_font_size=10, shadow=True, n_labels=9, italic=True, fmt="%.1f",
                         font_family="arial")
            if plot_type == "Clip":
                plot.add_text("Full Plot", font_size=15)
                if solution:
                    class MyCustomRoutine():
                        """ """
                        def __init__(self, mesh):
                            self.output = mesh  # Expected PyVista mesh type
                            # default parameters
                            self.kwargs = {
                                'min_val': 0.5,
                                'max_val': 30,
                            }

                        def __call__(self, param, value):
                            self.kwargs[param] = value
                            self.update()

                        def update(self):
                            """ """
                            # This is where you call your simulation
                            try:
                                plot.remove_actor("FieldPlot")
                            except:
                                pass
                            plot.add_mesh(surf, scalars=plot_label, log_scale=log, scalar_bar_args=sargs, cmap='rainbow',
                                          show_edges=False, clim=[self.kwargs['min_val'], self.kwargs['max_val']],
                                          pickable=True, smooth_shading=True, name="FieldPlot")
                            return

                    engine = MyCustomRoutine(surf)

                    plot.add_box_widget(surf, show_edges=False, scalars=plot_label, log_scale=log, scalar_bar_args=sargs,
                                        cmap='rainbow', pickable=True, smooth_shading=True, name="FieldPlot")

                    plot.add_slider_widget(callback=lambda value: engine('min_val', value),
                                           rng=[np.min(temps), np.max(temps)], title='Lower', style='modern',
                                           value=np.min(temps), pointa=(.5, .98), pointb=(.65, .98))

                    plot.add_slider_widget(callback=lambda value: engine('max_val', value),
                                           rng=[np.min(temps), np.max(temps)], title='Upper', style='modern',
                                           value=np.max(temps), pointa=(.66, .98), pointb=(.8, .98))
                else:
                    plot.add_box_widget(surf, show_edges=True, line_width=0.1, color="grey", pickable=True, smooth_shading=True)
            else:
                plot.add_text("Full Plot", font_size=15)
                if solution:
                    class MyCustomRoutine():
                        """ """
                        def __init__(self, mesh):
                            self.output = mesh  # Expected PyVista mesh type
                            # default parameters
                            self.kwargs = {
                                'min_val': 0.5,
                                'max_val': 30,
                            }

                        def __call__(self, param, value):
                            self.kwargs[param] = value
                            self.update()

                        def update(self):
                            """ """
                            # This is where you call your simulation
                            try:
                                plot.remove_actor("FieldPlot")
                            except:
                                pass
                            plot.add_mesh(surf, scalars=plot_label, log_scale=log, scalar_bar_args=sargs, cmap='rainbow',
                                          show_edges=False, clim=[self.kwargs['min_val'], self.kwargs['max_val']],
                                          pickable=True, smooth_shading=True, name="FieldPlot")
                            return

                    engine = MyCustomRoutine(surf)

                    plot.add_mesh(surf, show_edges=False, scalars=plot_label, log_scale=log, scalar_bar_args=sargs,
                                  cmap='rainbow', pickable=True, smooth_shading=True, name="FieldPlot")

                    plot.add_slider_widget(callback=lambda value: engine('min_val', value),
                                           rng=[np.min(temps), np.max(temps)], title='Lower', style='modern',
                                           value=np.min(temps), pointa=(.5, .98), pointb=(.65, .98))

                    plot.add_slider_widget(callback=lambda value: engine('max_val', value),
                                           rng=[np.min(temps), np.max(temps)], title='Upper', style='modern',
                                           value=np.max(temps), pointa=(.66, .98), pointb=(.8, .98))
                else:
                    plot.add_mesh(surf, show_edges=True, line_width=0.1, color="grey", pickable=True, smooth_shading=True)
            plot.show_axes()
            plot.show_grid()
            if view == "iso":
                plot.view_isometric()
            elif view == "x":
                plot.view_yz()
            elif view == "y":
                plot.view_xz()
            elif view == "z":
                plot.view_xy()
        files_list = []

        if plot:
            end = time.time() - start
            self._messenger.add_info_message("PyVista Generation tooks {} secs".format(end))
            if off_screen:
                if imageformat:
                    plot.show(screenshot=filename + "." + imageformat)
                    files_list.append(filename + "." + imageformat)
                else:
                    plot.show()
            else:
                def show(screen=None, interactive=True):
                    """

                    Parameters
                    ----------
                    screen : optional
                        The default is ``None``.
                    interactive : bool, optional
                        The default is ``True``.

                    Returns
                    -------

                    """
                    if screen:
                        plot.show(screenshot=screen, interactive=interactive, full_screen=True)
                    else:
                        plot.show(interactive=interactive)

                if imageformat:
                    show(filename + "." + imageformat, True)
                    files_list.append(filename + "." + imageformat)
                else:
                    show(filename + "." + imageformat, False)
            for f in aedtplt_files:
                os.remove(os.path.join(f))

        return files_list


    @aedt_exception_handler
    def _animation_from_aedtflt(self, aedtplt_files=None, variation_var="Time", variation_list=[],
                                plot_label="Temperature", model_color="#8faf8f", export_gif=False, off_screen=False):
        """Export the 3D field solver mesh, fields, or both mesh and fields as images using Python Plotly. 
        
          .. note::
           This function is currently supported only on Windows using CPython.

        Parameters
        ----------
        aedtplt_files : str or list, optional
            Names of the one or more AEDTPLT files generated by AEDT. The default is ``None``.
        variation_var : str, optional
             Variable to vary. The default is ``"Time"``.
        variation_list : list, optional
             List of variation values. The default is ``[]``.
        plot_label : str, optional
            Label for the plot. The default is ``"Temperature"``.
        model_color : str, optional
            Color scheme for the 3D model. The default is ``"#8faf8f"``, which is silver.
        export_gif : bool, optional
             Whether to export the animation as a GIF file. The default is ``False``.
        off_screen : bool, optional
             The default is ``False``.

        Returns
        -------
        str
            Name of the GIF file.
        """
        frame_per_seconds = 0.5
        start = time.time()
        if type(aedtplt_files) is str:
            aedtplt_files = [aedtplt_files]
        plot = pv.Plotter(notebook=False, off_screen=off_screen)
        if not off_screen:
            plot.enable_anti_aliasing()
        plot.enable_fly_to_right_click()
        lines = []
        for file in aedtplt_files:
            if ".aedtplt" in file:
                with open(file, "r") as f:
                    drawing_found = False
                    for line in f:
                        if "$begin Drawing" in line:
                            drawing_found = True
                            l_tmp = []
                            continue
                        if "$end Drawing" in line:
                            lines.append(l_tmp)
                            drawing_found = False
                            continue
                        if drawing_found:
                            l_tmp.append(line)
                            continue
                        if "Number of drawing:" in line:
                            n_drawings = int(line[18:])
                            continue
            elif ".obj" in file:
                mesh = pv.read(file)
                plot.add_mesh(mesh, show_scalar_bar=False, opacity=0.75, cmap=[model_color], name="3D Model",
                              show_edges=False, edge_color=model_color)
                # def create_object_mesh(opacity):
                #     try:
                #         p.remove_actor("Volumes")
                #     except:
                #         pass
                #     p.add_mesh(mesh, show_scalar_bar=False, opacity=opacity, cmap=[model_color], name="3D Model",
                #                show_edges=False, edge_color=model_color)
                # p.add_slider_widget(create_object_mesh, [0,1], style='modern', value=0.75,pointa=[0.81,0.98], pointb=[0.95,0.98],title="Opacity")
        filename = os.path.splitext(aedtplt_files[0])[0]
        print(filename)
        surfs=[]
        log = False
        mins=1e12
        maxs=-1e12
        log = True
        for drawing_lines in lines:
            bounding = []
            elements = []
            nodes_list = []
            solution = []
            for l in drawing_lines:
                if "BoundingBox(" in l:
                    bounding = l[l.find("(") + 1:-2].split(",")
                    bounding = [i.strip() for i in bounding]
                if "Elements(" in l:
                    elements = l[l.find("(") + 1:-2].split(",")
                    elements = [int(i.strip()) for i in elements]
                if "Nodes(" in l:
                    nodes_list = l[l.find("(") + 1:-2].split(",")
                    nodes_list = [float(i.strip()) for i in nodes_list]
                if "ElemSolution(" in l:
                    # convert list of strings to list of floats
                    sols = l[l.find("(") + 1:-2].split(",")
                    sols = [is_float(value) for value in sols]

                    num_solution_per_element = int(sols[2])
                    sols = sols[3:]
                    sols = [sols[i:i + num_solution_per_element] for i in range(0, len(sols), num_solution_per_element)]
                    solution = [sum(i) / num_solution_per_element for i in sols]

            nodes = [[nodes_list[i], nodes_list[i + 1], nodes_list[i + 2]] for i in range(0, len(nodes_list), 3)]
            num_nodes = elements[0]
            num_elements = elements[1]
            elements = elements[2:]
            element_type = elements[0]
            num_nodes_per_element = elements[4]
            hl = 5  # header length
            elements_nodes = []
            for i in range(0, len(elements), num_nodes_per_element + hl):
                elements_nodes.append([elements[i + hl + n] for n in range(num_nodes_per_element)])
            if solution:
                take_all_nodes = True  # solution case
            else:
                take_all_nodes = False  # mesh case
            trg_vertex = self._triangle_vertex(elements_nodes, num_nodes_per_element, take_all_nodes)
            # remove duplicates
            nodup_list = [list(i) for i in list(set([frozenset(t) for t in trg_vertex]))]
            sols_vertex = []
            if solution:
                sv = {}
                for els, s in zip(elements_nodes, solution):
                    for el in els:
                        if el in sv:
                            sv[el] = (sv[el] + s) / 2
                        else:
                            sv[el] = s
                sols_vertex = [sv[v] for v in sorted(sv.keys())]
            array = [[3] + [j - 1 for j in i] for i in nodup_list]
            faces = np.hstack(array)
            vertices = np.array(nodes)
            surf = pv.PolyData(vertices, faces)

            if sols_vertex:
                temps = np.array(sols_vertex)
                mean = np.mean(temps)
                std = np.std(temps)
                if np.min(temps) <=0:
                    log = False
                surf.point_arrays[plot_label] = temps
            if solution:
                surfs.append(surf)
                if np.min(temps)<mins:
                    mins=np.min(temps)
                if np.max(temps) > maxs:
                    maxs = np.max(temps)

        self._animating = True
        gifname=None
        if export_gif:
            gifname= os.path.splitext(aedtplt_files[0])[0]+".gif"
            plot.open_gif(gifname)
        def q_callback():
            """exit when user wants to leave"""
            self._animating = False
        self._pause=False
        def p_callback():
            """exit when user wants to leave"""
            self._pause = not self._pause
        plot.add_text('Press p for Play/Pause, Press q to exit ', font_size=8, position='upper_left')
        plot.add_text(' ', font_size=10, position=[0, 0])
        plot.add_key_event("q", q_callback)
        plot.add_key_event("p", p_callback)

        # run until q is pressed
        plot.show_axes()
        plot.show_grid()
        cpos = plot.show(interactive=False, auto_close=False,
                         interactive_update=not off_screen)


        sargs = dict(title_font_size=10, label_font_size=10,
                     shadow=True, n_labels=9, italic=True, fmt="%.1f",
                     font_family="arial")
        plot.add_mesh(surfs[0], scalars=plot_label, log_scale=log,
                      scalar_bar_args=sargs, cmap='rainbow',
                      clim=[mins, maxs], show_edges=False,
                      pickable=True, smooth_shading=True,
                      name="FieldPlot")
        plot.isometric_view()
        start = time.time()

        plot.update(1, force_redraw=True)
        first_loop = True
        if export_gif:
            first_loop = True
            plot.write_frame()
        else:
            first_loop = False
        i=1
        while self._animating:
            if self._pause:
                time.sleep(1)
                plot.update(1, force_redraw=True)
                continue
            #p.remove_actor("FieldPlot")
            if i >= len(surfs):
                if off_screen:
                    break
                i=0
                first_loop = False
            scalars = surfs[i].point_arrays[plot_label]
            plot.update_scalars(scalars, render=False)
            # p.add_mesh(surfs[i], scalars=plot_label, log_scale=log, scalar_bar_args=sargs, cmap='rainbow',
            #            show_edges=False, pickable=True, smooth_shading=True, name="FieldPlot")
            plot.textActor.SetInput(variation_var + " = " + variation_list[i])
            if not hasattr(plot, 'ren_win'):
                break
            #p.update(1, force_redraw=True)
            time.sleep(max(0, frame_per_seconds - (time.time() - start)))
            start = time.time()
            if off_screen:
                plot.render()
            else:
                plot.update(1, force_redraw=True)
            if first_loop:
                plot.write_frame()

            time.sleep(0.2)
            i+=1
        plot.close()
        for el in  aedtplt_files:
            os.remove(el)
        return gifname


    @aedt_exception_handler
    def export_model_obj(self):
        """Export the model."""
        assert self._parent._aedt_version >= "2021.2", self._messenger.add_error_message("Obj supported from AEDT 2021R2")
        project_path = self._parent.project_path
        obj_list = self._parent.modeler.primitives.object_names
        obj_list = [i for i in obj_list if not self._parent.modeler.primitives.objects[
            self._parent.modeler.primitives.get_obj_id(i)].is3d or (
                            self._parent.modeler.primitives.objects[
                                self._parent.modeler.primitives.get_obj_id(i)].material_name.lower() != "vacuum" and
                            self._parent.modeler.primitives.objects[
                                self._parent.modeler.primitives.get_obj_id(i)].material_name.lower() != "air")]
        self._parent.modeler.oeditor.ExportModelMeshToFile(os.path.join(project_path, "Model.obj"),
                                                           obj_list)
        return os.path.join(project_path, "Model.obj")

    @aedt_exception_handler
    def export_mesh_obj(self, setup_name=None, intrinsic_dict={}):
        """Export the mesh.

        Parameters
        ----------
        setup_name : str, optional
            Name of the setup. The default is ``None``.
        intrinsic_dict : dict, optipnal.
            Intrinsic dictionary needed for the export. 
            The default is ``{}``. 
          
        Returns
        -------

        """
        project_path = self._parent.project_path

        if not setup_name:
            setup_name = self._parent.nominal_adaptive
        face_lists = []
        obj_list = self._parent.modeler.primitives.object_names
        for el in obj_list:
            obj_id = self._parent.modeler.primitives.get_obj_id(el)
            if not self._parent.modeler.primitives.objects[obj_id].is3d or (
                    self._parent.modeler.primitives.objects[obj_id].material_name != "vacuum" and
                    self._parent.modeler.primitives.objects[obj_id].material_name != "air"):
                face_lists += self._parent.modeler.primitives.get_object_faces(obj_id)
        plot = self.create_fieldplot_surface(face_lists, "Mesh", setup_name, intrinsic_dict)
        if plot:
            file_to_add = self.export_field_plot(plot.name, project_path)
            plot.delete()
            return file_to_add
        return None

    @aedt_exception_handler
    def plot_model_obj(self, export_afterplot=True, jupyter=False):
        """Plot the model.

        Parameters
        ----------
        export_afterplot : bool, optional
             Whether to export the plot. The default is ``True``.
        jupyter : bool, optional
             Plot using jupyter.  The default is ``False``.

        Returns
        -------
        list
            List of AEDTPLT files.
        """
        assert self._parent._aedt_version >= "2021.2", self._messenger.add_error_message("Obj supported from AEDT 2021R2")
        files = [self.export_model_obj()]
        if export_afterplot:
            imageformat='jpg'
        else:
            imageformat=None
        file_list = self._plot_from_aedtplt(files, imageformat=imageformat, plot_label="3D Model", model_color="#8faf8f", show_model_edge=False, jupyter=jupyter)
        return file_list

    @aedt_exception_handler
    def plot_field_from_fieldplot(self, plotname, project_path="", meshplot=False, setup_name=None,
                                  intrinsic_dict={}, imageformat="jpg", view="iso", plot_label="Temperature", plot_folder=None, off_screen=False):
        """Export a field plot to an image file (JPG or PNG) using Python Plotly.
        
        .. note::
           The Plotly method rebuilds the mesh and overlap fields on the mesh.

        Parameters
        ----------
        plotname : str
            Name of the plot to export.
        project_path : str, optional
            Path where the image file is to be saved. The default is ``""``.
        meshplot : bool, optional
            Whether to create and plot the mesh over the fields. The
            default is ``False``.
        setup_name : str, optional
            Name of the setup or sweep to use for the export. The default is ``None``.
        intrinsic_dict : dict, optional
            Intrinsic dictionary needed for the export when ``meshplot="True"``. 
            The default is ``{}``. 
        imageformat : str, optional
            Format of the image file. Options are ``"jpg"``,
            ``"png"``, ``"svg"``, and ``"webp"``. The default is
            ``"jpg"``.
        view : str, optional
            View to export. Options are ``"iso"``, ``"x"`` , ``"y"``,
            ``"z"``, and ``"all"``.  The default is ``"iso"``. The
            ``"all"`` option exports all views.
        plot_label : str, optional
            Type of the plot. The default is ``"Temperature"``.
        plot_folder : str, optional
            Plot folder to forcibly update before exporting the
            field. The default is ``None``, which updates all of the
            plots.

        Returns
        -------
        type
            List of exported files.
        """
        if not plot_folder:
            self.ofieldsreporter.UpdateAllFieldsPlots()
        else:
            self.ofieldsreporter.UpdateQuantityFieldsPlots(plot_folder)

        start = time.time()
        files_to_add = []
        if not project_path:
            project_path = self._parent.project_path
        file_to_add = self.export_field_plot(plotname, project_path)
        file_list=None
        if not file_to_add:
            return False
        else:
            files_to_add.append(file_to_add)
            if meshplot:
                if self._parent._aedt_version >= "2021.2":
                    files_to_add.append(self.export_model_obj())
                else:
                    file_to_add = self.export_mesh_obj(setup_name, intrinsic_dict)
                    if file_to_add:
                        files_to_add.append(file_to_add)
            file_list = self._plot_from_aedtplt(files_to_add, imageformat=imageformat, view=view,
                                                plot_label=plot_label, off_screen=off_screen)
            endt = time.time() - start
            print("Field Generation, export and plot time: ", endt)
        return file_list

    @aedt_exception_handler
    def animate_fields_from_aedtplt(self, plotname, plot_folder=None, meshplot=False, setup_name=None,
                                    intrinsic_dict={}, variation_variable="Phi", variation_list=['0deg'],
                                    project_path="", export_gif=False, off_screen=False):
        """Generate a field plot to an image file (JPG or PNG) using PyVista.
        
        .. note::
           The PyVista method rebuilds the mesh and overlap fields on the mesh.

        Parameters
        ----------
        plotname : str
            Name of the plot or the name of the object.
        plot_folder : str, optional
            Name of the folder in which the plot resides. The default 
            is ``None``.
        setup_name : str, optional
            Name of the setup (sweep) to use for the export. The 
            default is ``None``.
        intrinsic_dict : dict, optional
            Intrinsic dictionary needed for the export. The default 
            is ``{}``. 
        variation_variable : str, optional
            Variable to vary. The default is ``"Phi"``.
        variation_list : list, optional
            List of variation values with units. The default is 
            ``['0deg']``.
        project_path : str, optional
            Path for the export. The default is ``""``.
        meshplot : bool, optional
             The default is ``False``.
        export_gif : bool, optional
             The default is ``False``.
        off_screen : bool, optional
             Generate the animation without showing an interactive plot.  The default is ``False``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        if not plot_folder:
            self.ofieldsreporter.UpdateAllFieldsPlots()
        else:
            self.ofieldsreporter.UpdateQuantityFieldsPlots(plot_folder)
        files_to_add = []
        if meshplot:
            if self._parent._aedt_version >= "2021.2":
                files_to_add.append(self.export_model_obj())
            else:
                file_to_add = self.export_mesh_obj(setup_name, intrinsic_dict)
                if file_to_add:
                    files_to_add.append(file_to_add)
        for el in variation_list:
            self._parent.odesign.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:FieldsPostProcessorTab",
                        [
                            "NAME:PropServers",
                            "FieldsReporter:"+plotname
                        ],
                        [
                            "NAME:ChangedProps",
                            [
                                "NAME:"+variation_variable,
                                "Value:="	, el
                            ]
                        ]
                    ]
                ])
            files_to_add.append(self.export_field_plot(plotname, project_path,plotname+variation_variable+str(el)))

        self._animation_from_aedtflt(files_to_add, variation_variable, variation_list, export_gif=export_gif, off_screen=off_screen)
        return True

    @aedt_exception_handler
    def animate_fields_from_aedtplt_2(self, quantityname, object_list, plottype, meshplot=False, setup_name=None,
                                    intrinsic_dict={}, variation_variable="Phi", variation_list=['0deg'],
                                    project_path="", export_gif=False, off_screen=False):
        """Generate a field plot to an image file (JPG or PNG) using PyVista.
        
         .. note::
            The PyVista method rebuilds the mesh and overlap fields on the mesh.
            
        This method creates the plot and exports it. It is alternative to the method ``animate_fields_from_aedtplt``, 
        which uses an existing plot.

        Parameters
        ----------
        quantityname : str
            Name of the plot or the name of the object.
        object_list : list, optional
            Name of the folderplot_folder
        plottype : str 
            Type of the plot. Options are ``"Surface"``, ``"Volume"``, and 
            ``"CutPlane"``.
        meshplot : bool, optional
            The default is ``False``.
        setup_name : str, optional
            Name of the setup (sweep) to use for the export. The default is 
            ``None``.
        intrinsic_dict : dict, optional
            Intrinsic dictionary needed for the export. 
            The default is ``{}``. 
        variation_variable : str, optional
            Variable to vary. The default is ``"Phi"``.
        variation_list : list, option
            List of variation values with units. The default is 
            ``['0deg']``.
        project_path : str, optional
            Path for the export. The default is ``""``.
        export_gif : bool, optional
             The default is ``False``.
        off_screen : bool, optional
             The default is ``False``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        if not project_path:
            project_path = self._parent.project_path
        files_to_add = []
        if meshplot:
            if self._parent._aedt_version >= "2021.2":
                files_to_add.append(self.export_model_obj())
            else:
                file_to_add = self.export_mesh_obj(setup_name, intrinsic_dict)
                if file_to_add:
                    files_to_add.append(file_to_add)
        v = 0
        for el in variation_list:
            intrinsic_dict[variation_variable] = el
            if plottype == "Surface":
                plotf = self.create_fieldplot_surface(object_list, quantityname, setup_name, intrinsic_dict)
            elif plottype == "Volume":
                plotf = self.create_fieldplot_volume(object_list, quantityname, setup_name, intrinsic_dict)
            else:
                plotf = self.create_fieldplot_cutplane(object_list, quantityname, setup_name, intrinsic_dict)
            if plotf:
                file_to_add = self.export_field_plot(plotf.name, project_path, plotf.name + str(v))
                if file_to_add:
                    files_to_add.append(file_to_add)
                plotf.delete()
            v += 1

        return self._animation_from_aedtflt(files_to_add, variation_variable, variation_list, export_gif=export_gif, off_screen=off_screen)


    @aedt_exception_handler
    def far_field_plot(self, ff_data, x=0, y=0, qty='rETotal', dB=True, array_size=[4, 4]):
        """

        Parameters
        ----------
        ff_data :
            
        x : float, optional
            The default is ``0``.
        y : float, optional
            The default is ``0``.
        qty : str, optional
            The default is ``"rETotal"``.
        dB : bool, optional
            The default is ``True``.
        array_size : list
            List for the array size. The default is ``[4, 4]``.          

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        loc_offset = 2  # if array index is not starting at [1,1]
        xphase = float(y)
        yphase = float(x)
        array_shape = (array_size[0], array_size[1])
        weight = np.zeros(array_shape, dtype=complex)
        mag = np.ones(array_shape, dtype='object')
        port_names_arranged = np.chararray(array_shape)
        all_ports = ff_data.keys()
        w_dict = {}
        # calculate weights based off of progressive phase shift
        port_name = []
        for m in range(array_shape[0]):
            for n in range(array_shape[1]):
                mag_val = mag[m][n]
                ang = np.radians(xphase * m) + np.radians(yphase * n)
                weight[m][n] = np.sqrt(mag_val) * np.exp(1j * ang)
                current_index_str = '[' + str(m + 1 + loc_offset) + ',' + str(n + 1 + loc_offset) + ']'
                port_name = [y for y in all_ports if current_index_str in y]
                w_dict[port_name[0]] = weight[m][n]

        length_of_ff_data = len(ff_data[port_name[0]][2])

        array_shape = (len(w_dict), length_of_ff_data)
        rEtheta_fields = np.zeros(array_shape, dtype=complex)
        rEphi_fields = np.zeros(array_shape, dtype=complex)
        w = np.zeros((1, array_shape[0]), dtype=complex)
        # create port mapping
        Ntheta = 0
        Nphi = 0
        for n, port in enumerate(ff_data.keys()):
            re_theta = ff_data[port][2]
            re_phi = ff_data[port][3]
            re_theta = re_theta * w_dict[port]

            w[0][n] = w_dict[port]
            re_phi = re_phi * w_dict[port]

            rEtheta_fields[n] = re_theta
            rEphi_fields[n] = re_phi

            theta_range = ff_data[port][0]
            phi_range = ff_data[port][1]
            theta = [int(np.min(theta_range)), int(np.max(theta_range)), np.size(theta_range)]
            phi = [int(np.min(phi_range)), int(np.max(phi_range)), np.size(phi_range)]
            Ntheta = len(theta_range)
            Nphi = len(phi_range)

        rEtheta_fields = np.dot(w, rEtheta_fields)
        rEtheta_fields = np.reshape(rEtheta_fields, (Ntheta, Nphi))

        rEphi_fields = np.dot(w, rEphi_fields)
        rEphi_fields = np.reshape(rEphi_fields, (Ntheta, Nphi))

        all_qtys = {}
        all_qtys['rEPhi'] = rEphi_fields
        all_qtys['rETheta'] = rEtheta_fields
        all_qtys['rETotal'] = np.sqrt(np.power(np.abs(rEphi_fields), 2) + np.power(np.abs(rEtheta_fields), 2))

        pin = np.sum(w)
        print(str(pin))
        real_gain = 2 * np.pi * np.abs(np.power(all_qtys['rETotal'], 2)) / pin / 377
        all_qtys['RealizedGain'] = real_gain

        if dB:
            if 'Gain' in qty:
                qty_to_plot = 10 * np.log10(np.abs(all_qtys[qty]))
            else:
                qty_to_plot = 20 * np.log10(np.abs(all_qtys[qty]))
            qty_str = qty + ' (dB)'
        else:
            qty_to_plot = np.abs(all_qtys[qty])
            qty_str = qty + ' (mag)'

        plt.figure(figsize=(15, 10))
        plt.title(qty_str)
        plt.xlabel('Theta (degree)')
        plt.ylabel('Phi (degree)')

        plt.imshow(qty_to_plot, cmap='jet')
        plt.colorbar()

        np.max(qty_to_plot)

    @aedt_exception_handler
    def create_3d_plot(self, solution_data, nominal_sweep="Freq", nominal_value=1, primary_sweep="Theta",
                       secondary_sweep="Phi"):
        """Create a 3D plot with Matplotlib.

        Parameters
        ----------
        solution_data :
            Input data for the solution.
        nominal_sweep : str, optional
            Name of the nominal sweep. The default is ``"Freq"``.
        nominal_value : str, optional
            Value for the nominal sweep. The default is ``1``.
        primary_sweep : str, optional
            Primary sweep. The default is ``"Theta"``.
        secondary_sweep : str, optional
            Secondary sweep. The default is ``"Phi"``.

        Returns
        -------
         bool
             ``True`` when successful, ``False`` when failed.
        """
        legend = []
        Freq = nominal_value
        solution_data.nominal_sweeps[nominal_sweep] = Freq
        solution_data.primary_sweep = primary_sweep
        solution_data.nominal_sweeps[primary_sweep] = 45
        theta = np.array((solution_data.sweeps[primary_sweep]))
        phi = np.array((solution_data.sweeps[secondary_sweep]))
        r = []
        i = 0
        phi1 = []
        theta1 = [i * math.pi / 180 for i in theta]
        for el in solution_data.sweeps[secondary_sweep]:
            solution_data.nominal_sweeps[secondary_sweep] = el
            phi1.append(el * math.pi / 180)
            r.append(solution_data.data_magnitude())
        THETA, PHI = np.meshgrid(theta1, phi1)

        R = np.array(r)
        X = R * np.sin(THETA) * np.cos(PHI)
        Y = R * np.sin(THETA) * np.sin(PHI)

        Z = R * np.cos(THETA)
        fig1 = plt.figure()
        ax1 = fig1.add_subplot(1, 1, 1, projection='3d')
        plot = ax1.plot_surface(
            X, Y, Z, rstride=1, cstride=1, cmap=plt.get_cmap('jet'),
            linewidth=0, antialiased=True, alpha=0.5)
        fig1.set_size_inches(10, 10)



import fnmatch
import os
import sys

# required since Python 2.7's glob.iglob does not support recursive keyword
def recursive_glob(startpath, filepattern):
    """Return a list of files matching a pattern, searching recursively from a start path.

    Keyword Arguments:
    startpath -- starting path (directory)
    filepattern -- fnmatch-style filename pattern
    """
    return [
        os.path.join(dirpath, filename)
            for dirpath, _, filenames in os.walk(startpath)
            for filename in filenames if fnmatch.fnmatch(filename, filepattern)
    ]

open3 = open
if sys.version_info < (3, 0):
    import io
    open3 = io.open

# loosely from https://packaging.python.org/guides/single-sourcing-package-version/
HERE = os.path.abspath(os.path.dirname(__file__))

name='pyaedt'

with open(os.path.join(HERE, 'pyaedt', 'version.txt'), 'r') as f:
    version = f.readline()

author='ANSYS, Inc.'

maintainer='Massimo Capodiferro'

maintainer_email='massimo.capodiferro@ansys.com'

description='Higher-Level Pythonic Ansys Electronics Destkop Framework'

# Get the long description from the README file
with open3(os.path.join(HERE, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

packages=['pyaedt', 'pyaedt.misc', 'pyaedt.application', 'pyaedt.modeler', 'pyaedt.modules',
            'pyaedt.generic', 'pyaedt.edb_core', 'pyaedt.examples']

data_files=[('dlls', recursive_glob(os.path.join('pyaedt', 'dlls'), '*')),
            ('misc', recursive_glob(os.path.join('pyaedt', 'misc'), '*')),
            ('License', recursive_glob('.', '*.md')),
            ('version', ['pyaedt/version.txt']),
            ('setup-distutils', ['setup-distutils.py'])]

license="MIT"

classifiers=[
    "Development Status :: 4 - Beta",
    "Programming Language :: Python :: 2.7",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "License :: OSI Approved :: MIT License",
    "Operating System :: Microsoft :: Windows",
    "Operating System :: POSIX",
]



"""
This module contains the ``EdbPadstacks`` class.
"""

import warnings
from .general import *
from ..generic.general_methods import get_filename_without_extension, generate_unique_name
from .EDB_Data import EDBPadstack
try:
    import clr
    from System import Convert, String
    from System import Double, Array
    from System.Collections.Generic import List
except ImportError:
    warnings.warn('This module requires pythonnet.')


class EdbPadstacks(object):
    """EdbPadstacks class."""

    def __init__(self, parent):
        self.parent = parent

    @property
    def _builder(self):
        """ """
        return self.parent.builder

    @property
    def _edb(self):
        """ """
        return self.parent.edb

    @property
    def _edb_value(self):
        """ """
        return self.parent.edb_value

    @property
    def _active_layout(self):
        """ """
        return self.parent.active_layout

    @property
    def db(self):
        """ """
        return self.parent.db

    @property
    def _padstack_methods(self):
        """ """
        return self.parent.edblib.Layout.PadStackMethods

    @property
    def _messenger(self):
        """ """
        return self.parent._messenger

    @property
    def _layers(self):
        """ """
        return self.parent.core_stackup.stackup_layers

    @property
    def padstacks(self):
        """List of padstacks via padstack definitions.
        
        Returns
        -------
        list
            List of padstacks via padstack definitions.

        """
        viadefList = {}
        for padstackdef in self.db.PadstackDefs:
            PadStackData = padstackdef.GetData()
            if len(PadStackData.GetLayerNames()) > 1:
                viadefList[padstackdef.GetName()] = EDBPadstack(padstackdef, self)
        return viadefList

    @aedt_exception_handler
    def create_circular_padstack(self, padstackname=None, holediam='300um', paddiam='400um', antipaddiam='600um',
                        startlayer=None, endlayer=None):
        """Create a new circular padstack.

        Parameters
        ----------
        padstackname : str, optional
            Name of the padstack. The default is ``None``.
        holediam : str, optional
            Diameter of the hole with units. The default is ``"300um"``.
        paddiam : str, optional
            Diameter of the pad with units. The default is ``"400um"``.
        antipaddiam : str, optional
            Diameter of the antipad with units. The default is ``"600um"``.
        startlayer : str, optional
            Starting layer. The default is ``None``, in which case the top
            is the starting layer.
        endlayer : str, optional
            Ending layer. The default is ``None``, in which case the bottom
            is the ending layer.
     
        Returns
        -------
        str
            Name of the padstack if the operation is successful.
        """
        return self._padstack_methods.CreateCircularPadStackDef(self._builder, padstackname, holediam, paddiam,
                                                               antipaddiam, startlayer, endlayer)

    @aedt_exception_handler
    def get_pinlist_from_component_and_net(self, refdes=None, netname=None):
        """Retrieve pins given a component's reference designator and net name.

        Parameters
        ----------
        refdes : str, optional
            Reference designator of the component. The default is ``None``.
        netname : str optional
            Name of the net. The default is ``None``.
        
        Returns
        -------
        dict
            Dictionary of pins if the operation is successful. 
            ``False`` is returned if the net does not belong to the component.

        """
        if self._builder:
            pinlist = self._padstack_methods.GetPinsFromComponentAndNets(self._builder, refdes, netname)
            if pinlist.Item1:
                return pinlist.Item2

    @aedt_exception_handler
    def create_padstack(self, padstackname=None, holediam='300um', paddiam='400um', antipaddiam='600um',
                        startlayer=None, endlayer=None):
        """Create a new padstack.

        Parameters
        ----------
        padstackname : str, optional
            Name of the padstack. The default is ``None``.
        holediam : str, optional
            Diameter of the hole with units. The default is ``"300um"``.
        paddiam : str, optional
            Diameter of the pad with units. The default is ``"400um"``.
        antipaddiam : str, optional
            Diameter of the antipad with units. The default is ``"600um"``.
        startlayer : str, optional
            Starting layer. The default is ``None``, in which case the top
            is the starting layer.
        endlayer : str, optional
            Ending layer. The default is ``None``, in which case the bottom
            is the ending layer.
            
        Returns
        -------
        str
            Name of the padstack if the operation is successful.
        """
        if not padstackname:
            padstackname = generate_unique_name("VIA")
        # assert not self.isreadonly, "Write Functions are not available within AEDT"
        padstackData = self._edb.Definition.PadstackDefData.Create()
        ptype = self._edb.Definition.PadGeometryType.Circle
        holparam = Array[type(self._edb_value(holediam))]([self._edb_value(holediam)])
        value0 = self._edb_value("0.0")

        padstackData.SetHoleParameters(ptype, holparam, value0, value0, value0)

        padstackData.SetHolePlatingPercentage(self._edb_value(20.0))
        padstackData.SetHoleRange(self._edb.Definition.PadstackHoleRange.UpperPadToLowerPad)
        padstackData.SetMaterial('copper')
        if not startlayer:
            layers = list(self.parent.core_stackup.signal_layers.keys())
            startlayer = layers[0]
        if not endlayer:
            layers = list(self.parent.core_stackup.signal_layers.keys())
            endlayer = layers[len(layers) - 1]
        for layer in [startlayer, 'Default', endlayer]:
            padparam_array = Array[type(self._edb_value(paddiam))]([self._edb_value(paddiam)])
            padstackData.SetPadParameters(
                layer,
                self._edb.Definition.PadType.RegularPad,
                self._edb.Definition.PadGeometryType.Circle,
                padparam_array,
                value0,
                value0,
                value0
            )
            antipad_array = Array[type(self._edb_value(antipaddiam))]([self._edb_value(antipaddiam)])

            padstackData.SetPadParameters(
                layer,
                self._edb.Definition.PadType.AntiPad,
                self._edb.Definition.PadGeometryType.Circle,
                antipad_array,
                value0,
                value0,
                value0
            )
        padstackLayerIdMap = {k: v for k, v in zip(padstackData.GetLayerNames(), padstackData.GetLayerIds())}
        padstackLayerMap = self._edb.Utility.LayerMap(self._edb.Utility.UniqueDirection.ForwardUnique)
        for layer, padstackLayerName in zip(
                self._active_layout.GetLayerCollection().Layers(
                    self._edb.Cell.LayerTypeSet.SignalLayerSet
                ),
                [startlayer, 'Default', endlayer]
        ):
            padstackLayerMap.SetMapping(
                layer.GetLayerId(),
                padstackLayerIdMap[padstackLayerName]
            )
        padstackDefinition = self._edb.Definition.PadstackDef.Create(self.db, padstackname)
        padstackDefinition.SetData(padstackData)
        self._messenger.add_info_message("Padstack {} create correctly".format(padstackname))
        return padstackname

    @aedt_exception_handler
    def place_padstack(self, position, definition_name, net_name='',
                       via_name="", rotation=0, fromlayer=None, tolayer=None, solderlayer=None):
        """Place the padstack.
        
        Parameters
        ----------
        position : list
            List of float values for the [x,y] positions where the via is to be placed. 
        definition_name : str
            Name of the padstack definition.
        net_name : str, optional
            Name of the net. The default is ``""``.
        via_name : str, optional
            The default is ``""``.
        rotation : float, optional
            Rotation of the padstack in degrees. The default  
            is ``0``.
        fromlayer :
            The default is ``None``.
        tolayer :
            The default is ``None``.
        solderlayer :
            The default is ``None``.
            
        Returns
        -------
        
        """    
        padstack = None
        for pad in list(self.padstacks.keys()):
            if pad == definition_name:
                padstack = self.padstacks[pad].edb_padstack
        position = self._edb.Geometry.PointData(self._edb_value(position[0]), self._edb_value(position[1]))
        net = self.parent.core_nets.find_or_create_net(net_name)
        rotation = self._edb_value(rotation)
        sign_layers = list(self.parent.core_stackup.signal_layers.keys())
        if not fromlayer:
            fromlayer = self.parent.core_stackup.signal_layers[sign_layers[-1]]._layer
        else:
            fromlayer = self.parent.core_stackup.signal_layers[fromlayer]._layer

        if not tolayer:
            tolayer = self.parent.core_stackup.signal_layers[sign_layers[0]]._layer
        else:
            tolayer = self.parent.core_stackup.signal_layers[tolayer]._layer
        if solderlayer:
            solderlayer = self.parent.core_stackup.signal_layers[solderlayer]._layer
        if padstack:
            via = self._edb.Cell.Primitive.PadstackInstance.Create(self._active_layout, net, via_name, padstack,
                                                                  position, rotation, fromlayer, tolayer, solderlayer,
                                                                  None)
            return via
        else:
            return False



"""
This module contains EDB general methods and related methods.

"""
from __future__ import absolute_import
import warnings


try:
    import clr
    clr.AddReference("System.Collections")
    from System.Collections.Generic import List
    from System import Int32
except ImportError:
    warnings.warn('This module requires pythonnet.')


import inspect
import itertools
import sys
import traceback
from collections import OrderedDict
from functools import wraps
import logging
logger = logging.getLogger(__name__)
from ..generic.general_methods import aedt_exception_handler, generate_unique_name

if "IronPython" in sys.version or ".NETFramework" in sys.version:
    is_ironpython = True
else:
    is_ironpython = False


@aedt_exception_handler
def convert_netdict_to_pydict(dict):
    """Convert a net dictionary to a Python dictionary.

    Parameters
    ----------
    dict : dict
        Net dictionary to convert.
    
    Returns
    -------
    dict
        Dictionary converted to Python.

    """
    pydict = {}
    for key in dict.Keys:
        pydict[key] = dict[key]
    return pydict

@aedt_exception_handler
def convert_pydict_to_netdict(dict):
    """Convert a Python dictionarty to a Net dictionary.

    Parameters
    ----------
    dict : dict
        Python dictionary to convert.
        

    Returns
    -------
    dict
        Dictionary converted to Net.
    """
    type = dict[dict.Keys[0]]
    # to be completed

@aedt_exception_handler
def convert_py_list_to_net_list(pylist):
    """Convert a Python list to a Net list.

    Parameters
    ----------
    pylist : list
        Python list to convert.
        
    Returns
    -------
    list
        List converted to Net.
    """
    if type(pylist) is not list and type(pylist) is not tuple:
        pylist = [pylist]
    ls = list([type(item) for item in pylist])
    if len(ls) > 0:
        net_list = List[ls[0]]()
        for el in pylist:
            net_list.Add(el)
        return net_list

@aedt_exception_handler
def convert_net_list_to_py_list(netlist):
    """Convert a Net list to a Python list.

    Parameters
    ----------
    netlist : list
       Net list to convert.
        

    Returns
    -------
    list
        List converted to Python.
    """
    pylist = []
    for el in netlist:
        pylist.__add__(el)
    return pylist
